#include <unistd.h>
#include <fcntl.h>

#include "pxll.h"

static int lookup_field (int tag, int label);

static
inline
pxll_int
get_typecode (object * ob)
{
  if (IMMEDIATE(ob)) {
    if (IS_INTEGER(ob)) {
      return TC_INT;
    } else {
      return (pxll_int)ob & 0xff;
    }
  } else {
    return (pxll_int)*((pxll_int *)ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case (object * ob)
{
  if (is_immediate (ob)) {
    if (is_int (ob)) {
      return TC_INT;
    } else {
      return (pxll_int)ob & 0xff;
    }
  } else {
    return (pxll_int)*((pxll_int *)ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case_noint (object * ob)
{
  if (is_immediate (ob)) {
    return (pxll_int) ob;
  } else {
    return (pxll_int) * ((pxll_int*) ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case_imm (object * ob)
{
  return (pxll_int)ob & 0xff;
}

static
inline
pxll_int
get_case_tup (object * ob)
{
  return (pxll_int)*((pxll_int *)ob) & 0xff;
}

static
inline
pxll_int
get_imm_payload (object * ob)
{
  return ((pxll_int) ob) >> 8;
}

static
pxll_int
get_tuple_size (object * ob)
{
  header * h = (header *) ob;
  return (*h)>>8;
}

static
void
indent (int n)
{
  while (n--) {
    fprintf (stdout, "  ");
  }
}

void print_string (object * ob, int quoted);
void print_list (pxll_pair * l);

// this is kinda lame, it's part pretty-printer, part not.
static
object *
dump_object (object * ob, int depth)
{
  // indent (depth);
  if (depth > 100) {
    fprintf (stdout , "...");
    return (object *) PXLL_UNDEFINED;
  }
  if (!ob) {
    fprintf (stdout, "<null>");
  } else if (is_int (ob)) {
    // integer
    fprintf (stdout, "%zd", unbox (ob));
  } else {
    int tc = is_immediate (ob);
    switch (tc) {
    case TC_CHAR:
      if ((pxll_int)ob>>8 == 257) {
	// deliberately out-of-range character
	fprintf (stdout, "#\\eof");
      } else {
	char ch = ((char)((pxll_int)ob>>8));
	switch (ch) {
	case ' ' : fprintf (stdout, "#\\space"); break;
	case '\n': fprintf (stdout, "#\\newline"); break;
	case '\r': fprintf (stdout, "#\\return"); break;
	case '\t': fprintf (stdout, "#\\tab"); break;
	default:   fprintf (stdout, "#\\%c", ch);
	}
      }
      break;
    case TC_BOOL:
      fprintf (stdout, ((pxll_int)ob >> 8) & 0xff ? "#t" : "#f");
      break;
    case TC_NIL:
      fprintf (stdout, "()");
      break;
    case TC_UNDEFINED:
      fprintf (stdout, "#u");
      break;
    case TC_EMPTY_VECTOR:
      fprintf (stdout, "#()");
      break;
    case 0: {
      // structure
      header h = (header) (ob[0]);
      int tc = h & 0xff;
      switch (tc) {
      case TC_SAVE: {
	// XXX fix me - now holds saved registers
        pxll_save * s = (pxll_save* ) ob;
        fprintf (stdout, "<save pc=%p\n", s->pc);
        dump_object ((object *) s->lenv, depth+1); fprintf (stdout, "\n");
        dump_object ((object *) s->next, depth+1); fprintf (stdout, ">");
      }
        break;
      case TC_CLOSURE: {
        pxll_closure * c = (pxll_closure *) ob;
        //fprintf (stdout, "<closure pc=%p\n", c->pc);
        //dump_object ((object *) c->lenv, depth+1); fprintf (stdout, ">\n");
	fprintf (stdout, "<closure pc=%p lenv=%p>", c->pc, c->lenv);
      }
        break;
      case TC_TUPLE: {
        pxll_tuple * t = (pxll_tuple *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "<tuple\n");
        for (i=0; i < n-1; i++) {
          dump_object ((object *) t->val[i], depth + 1); fprintf (stdout, "\n");
        }
        dump_object ((object *) t->next, depth + 1);
        fprintf (stdout, ">");
      }
	break;
      case TC_VECTOR: {
        pxll_vector * t = (pxll_vector *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "#(");
        for (i=0; i < n; i++) {
          dump_object ((object *) t->val[i], depth+1);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, ")");
      }
	break;
      case TC_VEC16: {
        pxll_vec16 * t = (pxll_vec16 *) ob;
        pxll_int n = t->len;
        int i;
	fprintf (stdout, "#16(");
        for (i=0; i < n; i++) {
	  fprintf (stdout, "%d", t->data[i]);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, ")");
      }
	break;
      case TC_PAIR:
	print_list ((pxll_pair *) ob);
        break;
      case TC_STRING:
	print_string (ob, 1);
	break;
      case TC_SYMBOL:
	print_string (ob[1], 0);
	break;
      default: {
        pxll_vector * t = (pxll_vector *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "{u%d ", (tc - TC_USEROBJ)>>2);
        for (i=0; i < n; i++) {
          dump_object ((object *) t->val[i], depth+1);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, "}");
      }
      }
    }
      break;
    default:
      // a user immediate unit-type...
      fprintf (stdout, "<u%d>", (tc>>8));
    }
  }
  return (object *) PXLL_UNDEFINED;
}

// for gdb...
void
DO (object * x)
{
  dump_object (x, 0);
  fprintf (stdout, "\n");
  fflush (stdout);
}

// for debugging
void
stack_depth_indent (object * k)
{
  while (k != PXLL_NIL) {
    k = k[1];
    fprintf (stderr, "  ");
  }
}

void
print_string (object * ob, int quoted)
{
  pxll_string * s = (pxll_string *) ob;
  char * ps = s->data;
  int i;
  //fprintf (stderr, "<printing string of len=%d (tuple-len=%d)>\n", s->len, get_tuple_size (ob));
  if (quoted) {
    fputc ('"', stdout);
  }
  for (i=0; i < (s->len); i++, ps++) {
    if (*ps == '"') {
      fputc ('\\', stdout);
      fputc ('"', stdout);
    } else {
      if (isprint(*ps)) {
	fputc (*ps, stdout);
      } else {
	fprintf (stdout, "\\0x%02x", *ps);
      }
    }
    if (i > 50) {
      fprintf (stdout, "...");
      break;
    }
  }
  if (quoted) {
    fputc ('"', stdout);
  }
}

void
print_list (pxll_pair * l)
{
  fprintf (stdout, "(");
  while (1) {
    object * car = l->car;
    object * cdr = l->cdr;
    dump_object (car, 0);
    if (cdr == PXLL_NIL) {
      fprintf (stdout, ")");
      break;
    } else if (!is_immediate (cdr) && GET_TYPECODE (*cdr) == TC_PAIR) {
      fprintf (stdout, " ");
      l = (pxll_pair *) cdr;
    } else {
      fprintf (stdout, " . ");
      dump_object (cdr, 0);
      fprintf (stdout, ")");
      break;
    }
  }
}

int
read_header (FILE * file)
{
  int depth = 0;
  // tiny lisp 'skipper' (as opposed to 'reader')
  do {
    char ch = fgetc (file);
    switch (ch) {
    case '(':
      depth++;
      break;
    case ')':
      depth--;
      break;
    case '"':
      while (fgetc (file) != '"') {
        // empty body
      }
      break;
    default:
      break;
    }
  } while (depth);
  // read terminating newline
  fgetc (file);
  return 0;
}

#ifndef NO_RANGE_CHECK
// used to check array references.  some day we might try to teach
//   the compiler when/how to skip doing this...
static
void
inline
range_check (unsigned int length, unsigned int index)
{
  if (index >= length) {
    fprintf (stderr, "array reference out of range\n");
    abort();
  }
}
#else
static
void
inline
range_check (unsigned int length, unsigned int index)
{
}
#endif

pxll_int verbose_gc = 1;
pxll_int clear_fromspace = 0;
pxll_int clear_tospace = 0;

pxll_int vm (int argc, char * argv[]);

#include "rdtsc.h"

unsigned long long gc_ticks = 0;

static
void
clear_space (object * p, pxll_int n)
{
  while (n--) {
    *p++ = PXLL_NIL;
  }
}

int
main (int argc, char * argv[])
{
  heap0 = malloc (sizeof (object) * heap_size);
  heap1 = malloc (sizeof (object) * heap_size);
  if (!heap0 || !heap1) {
    fprintf (stderr, "unable to allocate heap\n");
    return -1;
  } else {
    unsigned long long t0, t1;
    pxll_int result;
    if (clear_tospace) {
      clear_space (heap0, heap_size);
    }
    t0 = rdtsc();
    result = vm (argc, argv);
    t1 = rdtsc();
    dump_object ((object *) result, 0);
    fprintf (stdout, "\n");
    fprintf (stderr, "{total ticks: %lld gc ticks: %lld}\n", t1 - t0, gc_ticks);
    return (int) result;
  }
}

// CONSTRUCTED LITERALS //
pxll_string constructed_0 = { STRING_HEADER(12), 12, "lib/core.scm" };
pxll_string constructed_1 = { STRING_HEADER(1), 1, "}" };
pxll_string constructed_2 = { STRING_HEADER(1), 1, "}" };
pxll_string constructed_3 = { STRING_HEADER(2), 2, "{ " };
pxll_string constructed_4 = { STRING_HEADER(2), 2, "{ " };
pxll_string constructed_5 = { STRING_HEADER(1), 1, ")" };
pxll_string constructed_6 = { STRING_HEADER(1), 1, ")" };
pxll_string constructed_7 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_8 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_9 = { STRING_HEADER(2), 2, "#(" };
pxll_string constructed_10 = { STRING_HEADER(2), 2, "#(" };
pxll_string constructed_11 = { STRING_HEADER(2), 2, "#f" };
pxll_string constructed_12 = { STRING_HEADER(2), 2, "#f" };
pxll_string constructed_13 = { STRING_HEADER(2), 2, "#t" };
pxll_string constructed_14 = { STRING_HEADER(2), 2, "#t" };
pxll_string constructed_15 = { STRING_HEADER(1), 1, "'" };
pxll_string constructed_16 = { STRING_HEADER(1), 1, "'" };
pxll_string constructed_17 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_18 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_19 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_20 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_21 = { STRING_HEADER(5), 5, "colon" };
pxll_int constructed_22[] = {UPTR(22,1), SYMBOL_HEADER, UPTR0(21)};
pxll_string constructed_23 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_24 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_25 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_26 = { STRING_HEADER(1), 1, ":" };
pxll_string constructed_27 = { STRING_HEADER(5), 5, "quote" };
pxll_int constructed_28[] = {UPTR(28,1), SYMBOL_HEADER, UPTR0(27)};
pxll_string constructed_29 = { STRING_HEADER(1), 1, ")" };
pxll_string constructed_30 = { STRING_HEADER(1), 1, ")" };
pxll_string constructed_31 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_32 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_33 = { STRING_HEADER(1), 1, "(" };
pxll_string constructed_34 = { STRING_HEADER(1), 1, "(" };
pxll_string constructed_35 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_36 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_37 = { STRING_HEADER(1), 1, "=" };
pxll_string constructed_38 = { STRING_HEADER(1), 1, "=" };
pxll_string constructed_39 = { STRING_HEADER(0), 0, "" };
pxll_string constructed_40 = { STRING_HEADER(13), 13, "reading file " };
pxll_string constructed_41 = { STRING_HEADER(13), 13, "reading file " };
pxll_int constructed_42[] = {TC_NIL};
pxll_string constructed_43 = { STRING_HEADER(7), 7, "include" };
pxll_int constructed_44[] = {UPTR(44,1), SYMBOL_HEADER, UPTR0(43)};
pxll_int constructed_45[] = {TC_NIL};
pxll_string constructed_46 = { STRING_HEADER(30), 30, "expected '=' in record literal" };
pxll_int constructed_47[] = {TC_NIL};
pxll_int constructed_48[] = {TC_NIL};
pxll_string constructed_49 = { STRING_HEADER(30), 30, "expected closing ]/} character" };
pxll_string constructed_50 = { STRING_HEADER(11), 11, "%%array-ref" };
pxll_int constructed_51[] = {UPTR(51,1), SYMBOL_HEADER, UPTR0(50)};
pxll_string constructed_52 = { STRING_HEADER(22), 22, "unexpected end of file" };
pxll_int constructed_53[] = {TC_NIL};
pxll_string constructed_54 = { STRING_HEADER(30), 30, "bad backslash escape in string" };
pxll_string constructed_55 = { STRING_HEADER(5), 5, "comma" };
pxll_int constructed_56[] = {UPTR(56,1), SYMBOL_HEADER, UPTR0(55)};
pxll_string constructed_57 = { STRING_HEADER(3), 3, "nul" };
pxll_string constructed_58 = { STRING_HEADER(3), 3, "eof" };
pxll_string constructed_59 = { STRING_HEADER(3), 3, "tab" };
pxll_string constructed_60 = { STRING_HEADER(6), 6, "return" };
pxll_string constructed_61 = { STRING_HEADER(5), 5, "space" };
pxll_string constructed_62 = { STRING_HEADER(7), 7, "newline" };
pxll_string constructed_63 = { STRING_HEADER(22), 22, "bad character constant" };
pxll_string constructed_64 = { STRING_HEADER(12), 12, "syntax error" };
pxll_string constructed_65 = { STRING_HEADER(22), 22, "unexpected close-paren" };
pxll_string constructed_66 = { STRING_HEADER(18), 18, "bad decimal digit?" };
pxll_int constructed_67[] = {TC_NIL};
pxll_string constructed_68 = { STRING_HEADER(52), 52, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" };
pxll_int constructed_69[] = {UPTR(69,37), UOHEAD(3,1), 14594, 19, UITAG(0), UOHEAD(3,1), 14338, 17, UPTR(69,1), UOHEAD(3,1), 14082, 15, UPTR(69,5), UOHEAD(3,1), 13826, 13, UPTR(69,9), UOHEAD(3,1), 13570, 11, UPTR(69,13), UOHEAD(3,1), 13314, 9, UPTR(69,17), UOHEAD(3,1), 13058, 7, UPTR(69,21), UOHEAD(3,1), 12802, 5, UPTR(69,25), UOHEAD(3,1), 12546, 3, UPTR(69,29), UOHEAD(3,1), 12290, 1, UPTR(69,33)};
pxll_int constructed_70[] = {UPTR(70,10), UOHEAD(2,-3), 3330, TC_NIL, UOHEAD(2,-3), 2562, UPTR(70,1), UOHEAD(2,-3), 2306, UPTR(70,4), UOHEAD(2,-3), 8194, UPTR(70,7)};
pxll_string constructed_71 = { STRING_HEADER(10), 10, "0123456789" };
pxll_string constructed_72 = { STRING_HEADER(7), 7, "()[]{}:" };
pxll_string constructed_73 = { STRING_HEADER(13), 13, "bad hex digit" };
pxll_int constructed_74[] = {UPTR(74,85), UOHEAD(3,1), 17922, 31, UITAG(0), UOHEAD(3,1), 17666, 29, UPTR(74,1), UOHEAD(3,1), 17410, 27, UPTR(74,5), UOHEAD(3,1), 17154, 25, UPTR(74,9), UOHEAD(3,1), 16898, 23, UPTR(74,13), UOHEAD(3,1), 16642, 21, UPTR(74,17), UOHEAD(3,1), 26114, 31, UPTR(74,21), UOHEAD(3,1), 25858, 29, UPTR(74,25), UOHEAD(3,1), 25602, 27, UPTR(74,29), UOHEAD(3,1), 25346, 25, UPTR(74,33), UOHEAD(3,1), 25090, 23, UPTR(74,37), UOHEAD(3,1), 24834, 21, UPTR(74,41), UOHEAD(3,1), 14594, 19, UPTR(74,45), UOHEAD(3,1), 14338, 17, UPTR(74,49), UOHEAD(3,1), 14082, 15, UPTR(74,53), UOHEAD(3,1), 13826, 13, UPTR(74,57), UOHEAD(3,1), 13570, 11, UPTR(74,61), UOHEAD(3,1), 13314, 9, UPTR(74,65), UOHEAD(3,1), 13058, 7, UPTR(74,69), UOHEAD(3,1), 12802, 5, UPTR(74,73), UOHEAD(3,1), 12546, 3, UPTR(74,77), UOHEAD(3,1), 12290, 1, UPTR(74,81)};
pxll_int constructed_75[] = {UPTR(75,10), UOHEAD(2,-3), 3330, TC_NIL, UOHEAD(2,-3), 2562, UPTR(75,1), UOHEAD(2,-3), 2306, UPTR(75,4), UOHEAD(2,-3), 8194, UPTR(75,7)};
pxll_string constructed_76 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_77 = { STRING_HEADER(1), 1, " " };
pxll_string constructed_78 = { STRING_HEADER(13), 13, "open() failed" };
pxll_string constructed_79 = { STRING_HEADER(13), 13, "open() failed" };
pxll_int pxll_internal_symbols[] = {(5<<8)|TC_VECTOR, UPTR(28,1), UPTR(44,1), UPTR(22,1), UPTR(56,1), UPTR(51,1)};

pxll_int
vm (int argc, char * argv[])
{
  register object * lenv = PXLL_NIL;
  register object * k = PXLL_NIL;
// REGISTER_DECLARATIONS //
  register object * r0;
  register object * r1;
  register object * r2;
  register object * r3;
  register object * r4;
  register object * r5;
  register object * r6;
  register object * r7;
  register object * r8;
  register object * r9;
  register object * r10;
  register object * r11;
  register object * r12;
  register object * r13;  
  void gc_regs_in (int n) {
    switch (n) {
    case 14: heap1[16] = r13;
    case 13: heap1[15] = r12;
    case 12: heap1[14] = r11;
    case 11: heap1[13] = r10;
    case 10: heap1[12] = r9;
    case 9: heap1[11] = r8;
    case 8: heap1[10] = r7;
    case 7: heap1[9] = r6;
    case 6: heap1[8] = r5;
    case 5: heap1[7] = r4;
    case 4: heap1[6] = r3;
    case 3: heap1[5] = r2;
    case 2: heap1[4] = r1;
    case 1: heap1[3] = r0;
  }}
  void gc_regs_out (int n) {
    switch (n) {
    case 14: r13 = heap0[16];
    case 13: r12 = heap0[15];
    case 12: r11 = heap0[14];
    case 11: r10 = heap0[13];
    case 10: r9 = heap0[12];
    case 9: r8 = heap0[11];
    case 8: r7 = heap0[10];
    case 7: r6 = heap0[9];
    case 6: r5 = heap0[8];
    case 5: r4 = heap0[7];
    case 4: r3 = heap0[6];
    case 3: r2 = heap0[5];
    case 2: r1 = heap0[4];
    case 1: r0 = heap0[3];
  }}
  object * top = PXLL_NIL; // top-level (i.e. 'global') environment
  object * t = 0; // temp - for swaps & building tuples
  object * result;
  object * limit = heap0 + (heap_size - head_room);
  object * freep = heap0;
  int i; // loop counter
  
#define PXLL_RETURN(d)	result = r##d; goto *k[3]

#include "gc.c"

  // check heap is called at the top of each allocating function.
  //  [by locating the check at the top, we avoid considering any
  //   registers as roots of the gc...]
  void check_heap (int nfree) {
    if (freep >= limit) {
      uint64_t t0, t1;
      t0 = rdtsc();
      gc_flip (nfree);
      t1 = rdtsc();
      gc_ticks += t1 - t0;
    }
  }

  object * allocate (pxll_int tc, pxll_int size) {
    object * save = freep;
    *freep = (object*) (size<<8 | (tc & 0xff));
#if 1
    // at least on the g5, this technique is considerably faster than using memset
    //   in gc_flip() to 'pre-clear' the heap... probably a cache effect...
    while (size--) {
      // this keeps gc from being confused by partially-filled objects.
      *(++freep) = PXLL_NIL;
    }
    ++freep;
#else
    // if you use this version, be sure to set <clear_tospace>!
    freep += size + 1;
#endif
    return save;  
  }

  // this is emitted by the backend for %make-tuple
  object * alloc_no_clear (pxll_int tc, pxll_int size) {
    object * save = freep;
    *freep = (object*) (size<<8 | (tc & 0xff));
    freep += size + 1;
    return save;  
  }

  // gcc inlines/unrolls these nicely, they allow more compact code
  inline object * varref (pxll_int depth, pxll_int index)
  {
    object * walk = lenv;
    while (depth--) {
      walk = walk[1];
    }
    return walk[index+2];
  }

  inline void varset (pxll_int depth, pxll_int index, object * val)
  {
    object * walk = lenv;
    while (depth--) {
      walk = walk[1];
    }
    walk[index+2] = val;
  }

  
  // these could probably be written in irken...
  pxll_int dump_image (char * filename, object * closure) {
    FILE * dump_file = fopen (filename, "wb");
    pxll_int offset;
    pxll_int size;
    object * start;
    // do a gc for a compact dump
    closure = gc_dump (closure);
    // for now, start at the front of the heap
    start = heap0;
    size = freep - start;
    offset = (pxll_int) heap0;
    // XXX add endian indicator...
    fprintf (dump_file, "(pxll image %ld %p)\n", sizeof (pxll_int), start);
    fwrite (&offset, sizeof(pxll_int), 1, dump_file);
    fwrite (&size, sizeof(pxll_int), 1, dump_file);
    fwrite (start, sizeof(pxll_int), size, dump_file);
    fclose (dump_file);
    return size;
  }

  object * load_image (char * filename) {
    FILE * load_file = fopen (filename, "rb");
    if (!load_file) {
      abort();
    } else {
      object * start, * thunk;
      pxll_int size;
      read_header (load_file);	// XXX verify header...
      fread (&start, sizeof(pxll_int), 1, load_file);
      fread (&size, sizeof(pxll_int), 1, load_file);
      fread (heap1, sizeof(pxll_int), size, load_file);
      fprintf (stderr, "size=%d\n", (int) size);
      // relocate heap0
      gc_relocate (4, heap1, heap1 + size, start - heap1);
      // replace roots
      lenv  = (object *) heap1[0];
      k     = (object *) heap1[1];
      top   = (object *) heap1[2];
      thunk = (object *) heap1[3];
      freep = heap1 + size;
      // swap heaps
      { object * temp = heap0; heap0 = heap1; heap1 = temp; }
      return thunk;
    }
  }

  k = allocate (TC_SAVE, 3);
  k[1] = (object *) PXLL_NIL; // top of stack
  k[2] = (object *) PXLL_NIL; // null environment
  k[3] = &&Lreturn; // continuation that will return from this function.
  // --- BEGIN USER PROGRAM ---
  r0 = allocate (TC_TUPLE, 22);
  top = r0;
  r0[1] = lenv; lenv = r0;
  // def 'char-class_131'
  goto L1;
  FUN_char_class_131:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) PXLL_FALSE;
    r0[2] = r1;
    r1 = (object *) 513;
    r2 = varref (0,0);
    if (unbox(r1) == 0) { r1 = (object *) TC_EMPTY_VECTOR; } else {
      t = alloc_no_clear (TC_VECTOR, unbox(r1));
      for (i=0; i < unbox(r1); i++) { t[i+1] = r2; }
      r1 = t;
    }
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def 'in-class?_540'
    goto L2;
    FUN_in_class_question_540:
      r0 = varref (0,0);
      r0 = box(GET_CHAR(r0));
      r1 = varref (2,1);
      range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r0));
      r0 = ((pxll_vector*)r1)->val[unbox(r0)];
      PXLL_RETURN(0);
    L2:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_in_class_question_540; r1[2] = lenv;
    r0[2] = r1;
    // def 'loop_542'
    goto L3;
    FUN_loop_542:
      r0 = varref (0,0);
      switch (get_case_noint (r0)) {
        case (((pxll_int)TC_PAIR)): {
          r0 = varref (0,0);
          r0 = UOBJ_GET(r0,0);
          r1 = varref (0,0);
          r1 = UOBJ_GET(r1,1);
          r2 = r0;
          r2 = box(GET_CHAR(r2));
          r3 = varref (2,1);
          r4 = (object *) PXLL_TRUE;
          range_check (GET_TUPLE_LENGTH(*(object*)r3), unbox(r2));
          ((pxll_vector*)r3)->val[unbox(r2)] = r4;
          r2 = r1;
          lenv[2] = r2;
          goto FUN_loop_542;
        } break;
        default: {
          r0 = varref (1,0);
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L3:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_542; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = varref (2,0);
    r0[2] = r1;
    r1 = varref (0,1);
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_542;
  L1:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_char_class_131; r1[2] = lenv;
  r0[2] = r1;
  // def 'file/fill-buffer_113'
  goto L4;
  FUN_file_2ffill_buffer_113:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 6);
    r0[1] = lenv; lenv = r0;
    r1 = varref (1,0);
    r1 = ((pxll_vector*)r1)->val[2];
    r0[2] = r1;
    r1 = varref (1,0);
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = varref (0,1);
    r1 = box(((pxll_string*)(r1))->len);
    r0[4] = r1;
    r1 = varref (0,0);
    r2 = varref (0,1);
    r3 = varref (0,2);
    r1 = box(read (unbox(r1), ((pxll_string*)(r2))->data, unbox(r3)));
    r0[5] = r1;
    r1 = varref (0,3);
    r0[6] = r1;
    r0 = varref (1,0);
    r1 = varref (0,4);
    ((pxll_vector*)r0)->val[1] = r1;
    r0 = varref (1,0);
    r1 = (object *) 1;
    ((pxll_vector*)r0)->val[3] = r1;
    r0 = varref (0,4);
    PXLL_RETURN(0);
  L4:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2ffill_buffer_113; r1[2] = lenv;
  r0[3] = r1;
  // def 'file/read-char_115'
  goto L5;
  FUN_file_2fread_char_115:
    check_heap (0);
    r0 = varref (0,0);
    r0 = ((pxll_vector*)r0)->val[3];
    r1 = varref (0,0);
    r1 = ((pxll_vector*)r1)->val[1];
    r2 = r0;
    r3 = r1;
    r0 = PXLL_TEST(unbox(r2)<unbox(r3));
    if PXLL_IS_TRUE(r0) {
      r0 = varref (0,0);
      r0 = ((pxll_vector*)r0)->val[3];
      r1 = r0;
      r2 = (object *) 3;
      r0 = box(unbox(r1)+unbox(r2));
      r1 = varref (0,0);
      ((pxll_vector*)r1)->val[3] = r0;
      r0 = varref (0,0);
      r0 = ((pxll_vector*)r0)->val[0];
      r1 = varref (0,0);
      r1 = ((pxll_vector*)r1)->val[3];
      r2 = r1;
      r3 = (object *) 3;
      r2 = box(unbox(r2)-unbox(r3));
      r3 = r0;
      r4 = r2;
      r3 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r4)]);
      PXLL_RETURN(3);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = varref (1,0);
      r1[2] = r2;
      r2 = top[3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L6; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_file_2ffill_buffer_113;
      L6:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (0,0);
      r0 = PXLL_TEST(unbox(r0)==0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 65794;
        PXLL_RETURN(0);
      } else {
        r0 = varref (0,0);
        lenv[2] = r0;
        goto FUN_file_2fread_char_115;
      }
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L5:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2fread_char_115; r1[2] = lenv;
  r0[4] = r1;
  // def 'reverse-onto_69'
  goto L7;
  FUN_reverse_onto_69:
    check_heap (0);
    r0 = varref (0,0);
    switch (get_case_noint (r0)) {
      case (((pxll_int)TC_NIL)): {
        r0 = varref (0,1);
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = varref (0,0);
        r0 = UOBJ_GET(r0,0);
        r1 = varref (0,0);
        r1 = UOBJ_GET(r1,1);
        r2 = r0;
        r3 = varref (0,1);
        t = alloc_no_clear (TC_PAIR, 2);
        t[1] = r2; t[2] = r3;
        r2 = t;
        r3 = r1;
        lenv[2] = r3;
        lenv[3] = r2;
        goto FUN_reverse_onto_69;
      } break;
    }
    PXLL_RETURN(0);
  L7:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_reverse_onto_69; r1[2] = lenv;
  r0[5] = r1;
  // def 'alist/lookup_118'
  goto L8;
  FUN_alist_2flookup_118:
    check_heap (0);
    r0 = varref (0,0);
    switch (get_case_noint (r0)) {
      case (TC_USEROBJ+4): {
        r0 = varref (0,0);
        r0 = UOBJ_GET(r0,0);
        r1 = varref (0,0);
        r1 = UOBJ_GET(r1,1);
        r2 = varref (0,0);
        r2 = UOBJ_GET(r2,2);
        r3 = r0;
        r4 = varref (0,1);
        if PXLL_IS_TRUE(PXLL_TEST(r3==r4)) {
          r3 = r1;
          t = alloc_no_clear ((TC_USEROBJ+0), 1);
          t[1] = r3;
          r3 = t;
          PXLL_RETURN(3);
        } else {
          r3 = r2;
          r4 = varref (0,1);
          lenv[2] = r3;
          lenv[3] = r4;
          goto FUN_alist_2flookup_118;
        }
        PXLL_RETURN(3);
      } break;
      default: {
        r0 = (object*)UITAG(1);
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L8:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist_2flookup_118; r1[2] = lenv;
  r0[6] = r1;
  // def 'length_72'
  goto L9;
  FUN_length_72:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def 'loop_262'
    goto L10;
    FUN_loop_262:
      r0 = varref (0,0);
      switch (get_case_noint (r0)) {
        case (((pxll_int)TC_NIL)): {
          r0 = varref (0,1);
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = varref (0,0);
          r0 = UOBJ_GET(r0,1);
          r1 = varref (0,1);
          r2 = (object *) 3;
          r1 = box(unbox(r1)+unbox(r2));
          r2 = r0;
          lenv[2] = r2;
          lenv[3] = r1;
          goto FUN_loop_262;
        } break;
      }
      PXLL_RETURN(0);
    L10:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_262; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = varref (1,0);
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = varref (0,0);
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_262;
  L9:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_length_72; r1[2] = lenv;
  r0[7] = r1;
  // def 'list->string_97'
  goto L11;
  FUN_list__3estring_97:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (1,0);
    r1[2] = r2;
    r2 = top[7];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L12; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_72;
    L12:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = varref (0,0);
    r2 = varref (0,0);
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r2), t);
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def 'loop_369'
    goto L13;
    FUN_loop_369:
      r0 = varref (0,0);
      switch (get_case_noint (r0)) {
        case (((pxll_int)TC_PAIR)): {
          r0 = varref (0,0);
          r0 = UOBJ_GET(r0,0);
          r1 = varref (0,0);
          r1 = UOBJ_GET(r1,1);
          r2 = varref (2,1);
          r3 = varref (0,1);
          r4 = r0;
          (((pxll_string*)(r2))->data[unbox(r3)] = GET_CHAR (r4), PXLL_UNDEFINED);
          r2 = varref (0,1);
          r3 = (object *) 3;
          r2 = box(unbox(r2)+unbox(r3));
          r3 = r1;
          lenv[2] = r3;
          lenv[3] = r2;
          goto FUN_loop_369;
        } break;
        default: {
          r0 = varref (2,1);
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L13:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_369; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = varref (2,0);
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = varref (0,0);
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_369;
  L11:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_list__3estring_97; r1[2] = lenv;
  r0[8] = r1;
  // def 'string->list_98'
  goto L14;
  FUN_string__3elist_98:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def 'loop_375'
    goto L15;
    FUN_loop_375:
      check_heap (0);
      r0 = varref (0,1);
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r0)==0)) {
        r0 = varref (0,0);
        PXLL_RETURN(0);
      } else {
        r0 = varref (0,1);
        r1 = (object *) 3;
        r0 = box(unbox(r0)-unbox(r1));
        r1 = varref (2,0);
        r2 = r0;
        r0 = TO_CHAR(((unsigned char *)((pxll_string*)(r1))->data)[unbox(r2)]);
        r1 = varref (0,0);
        t = alloc_no_clear (TC_PAIR, 2);
        t[1] = r0; t[2] = r1;
        r0 = t;
        r1 = varref (0,1);
        r2 = (object *) 3;
        r1 = box(unbox(r1)-unbox(r2));
        lenv[2] = r0;
        lenv[3] = r1;
        goto FUN_loop_375;
      }
      PXLL_RETURN(0);
    L15:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_375; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) TC_NIL;
    r0[2] = r1;
    r1 = varref (1,0);
    r1 = box(((pxll_string*)(r1))->len);
    r0[3] = r1;
    r1 = varref (0,0);
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_375;
  L14:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string__3elist_98; r1[2] = lenv;
  r0[9] = r1;
  r1 = (object*)UITAG(2);
  r0[10] = r1;
  // def 'zero-terminate_96'
  goto L16;
  FUN_zero_terminate_96:
    r0 = varref (0,0);
    r0 = box(((pxll_string*)(r0))->len);
    r1 = r0;
    r2 = (object *) 3;
    r1 = box(unbox(r1)-unbox(r2));
    r2 = varref (0,0);
    r3 = r1;
    r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r3)]);
    r3 = r2;
    r4 = (object *) 2;
    r0 = PXLL_TEST(r3==r4);
    if PXLL_IS_TRUE(r0) {
      r0 = varref (0,0);
      PXLL_RETURN(0);
    } else {
      r0 = varref (0,0);
      r0 = box(((pxll_string*)(r0))->len);
      r1 = r0;
      r2 = (object *) 3;
      r1 = box(unbox(r1)+unbox(r2));
      r2 = r1;
      r3 = r1;
      r2 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r2))), ((pxll_string*)(t))->len = unbox(r3), t);
      r3 = r2;
      r4 = (object *) 1;
      r5 = varref (0,0);
      r6 = (object *) 1;
      r7 = r0;
      (memcpy (((pxll_string*)(r3))->data+unbox(r4), ((pxll_string*)(r5))->data+unbox(r6), unbox(r7)), PXLL_UNDEFINED);
      r3 = r2;
      r4 = r0;
      r5 = (object *) 2;
      (((pxll_string*)(r3))->data[unbox(r4)] = GET_CHAR (r5), PXLL_UNDEFINED);
      r3 = r2;
      PXLL_RETURN(3);
    }
    PXLL_RETURN(0);
  L16:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_zero_terminate_96; r1[2] = lenv;
  r0[11] = r1;
  // def 'open_101'
  goto L17;
  FUN_open_101:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (1,0);
    r1[2] = r2;
    r2 = top[11];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L18; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_96;
    L18:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = varref (0,0);
    r0 = dump_object (r0, 0); fprintf (stdout, "\n");
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (1,0);
    r1[2] = r2;
    r2 = top[11];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L19; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_96;
    L19:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r2 = varref (1,1);
    r3 = varref (1,2);
    r1 = box(open (((pxll_string*)(r1))->data, unbox(r2), unbox(r3)));
    r0[2] = r1;
    r0 = varref (0,0);
    r1 = (object *) 1;
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r0)>=unbox(r1))) {
      r0 = varref (0,0);
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = varref (2,0);
      r1[2] = r2;
      r2 = top[11];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L20; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_96;
      L20:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 2562;
      box(fputc (GET_CHAR(r0), stdout));
      r0 = (object*) &constructed_79;
      r0 = box(((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_78;
      box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
      r0 = (object*) &constructed_77;
      r0 = box(((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_76;
      box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
      r0 = varref (0,0);
      dump_object (r0, 0); fprintf (stdout, "\n");
      goto Lreturn;
      r0 = PXLL_UNDEFINED;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L17:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_open_101; r1[2] = lenv;
  r0[12] = r1;
  // def 'string-compare_92'
  goto L21;
  FUN_string_compare_92:
    r0 = varref (0,0);
    r0 = box(((pxll_string*)(r0))->len);
    r1 = varref (0,1);
    r1 = box(((pxll_string*)(r1))->len);
    r2 = r0;
    r3 = r1;
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r3))) {
      r2 = r0;
    } else {
      r2 = r1;
    }
    r3 = varref (0,0);
    r4 = varref (0,1);
    r2 = box(memcmp (((pxll_string*)(r3))->data, ((pxll_string*)(r4))->data, unbox(r2)));
    r3 = r2;
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)==0)) {
      r3 = r0;
      r4 = r1;
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)==unbox(r4))) {
        r3 = (object *) 1;
        PXLL_RETURN(3);
      } else {
        r3 = r0;
        r4 = r1;
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)<unbox(r4))) {
          r3 = (object *) -1;
          PXLL_RETURN(3);
        } else {
          r3 = (object *) 3;
          PXLL_RETURN(3);
        }
        PXLL_RETURN(3);
      }
      PXLL_RETURN(3);
    } else {
      r3 = r2;
      PXLL_RETURN(3);
    }
    PXLL_RETURN(3);
  L21:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_compare_92; r1[2] = lenv;
  r0[13] = r1;
  // def 'intern-symbol_127'
  goto L22;
  FUN_intern_symbol_127:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = top[10];
    r0[2] = r1;
    r1 = varref (1,0);
    r1 = varref (1,0);
    r1 = UOBJ_GET(r1,0);
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def 'ins_439_i52'
    goto L23;
    FUN_ins_439_i52:
      check_heap (0);
      r0 = varref (0,0);
      switch (get_case_noint (r0)) {
        case (TC_USEROBJ+4): {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,3);
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = varref (4,1);
          r1[2] = r2;
          r2 = varref (1,2);
          r1[3] = r2;
          r2 = top[13];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L24; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
          L24:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = varref (0,0);
          r0 = PXLL_TEST(unbox(r0)<0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (0,0);
            r0[2] = r1;
            r1 = varref (2,0);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L25;  k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_ins_439_i52;
            L25:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            r1 = r0;
            switch (get_case_noint (r1)) {
              case (TC_USEROBJ+0): {
                r1 = r0;
                r1 = UOBJ_GET(r1,0);
                r2 = r0;
                r2 = UOBJ_GET(r2,1);
                r3 = r0;
                r3 = UOBJ_GET(r3,2);
                r4 = r0;
                r4 = UOBJ_GET(r4,3);
                r5 = r1;
                switch (get_case_noint (r5)) {
                  case (TC_USEROBJ+0): {
                    r5 = r1;
                    r5 = UOBJ_GET(r5,0);
                    r6 = r1;
                    r6 = UOBJ_GET(r6,1);
                    r7 = r1;
                    r7 = UOBJ_GET(r7,2);
                    r8 = r1;
                    r8 = UOBJ_GET(r8,3);
                    r9 = r5;
                    r10 = r6;
                    r11 = r7;
                    r12 = r8;
                    t = alloc_no_clear ((TC_USEROBJ+4), 4);
                    t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                    r9 = t;
                    r10 = r2;
                    r11 = varref (0,1);
                    r12 = varref (0,2);
                    r13 = varref (0,3);
                    t = alloc_no_clear ((TC_USEROBJ+4), 4);
                    t[1] = r10; t[2] = r11; t[3] = r12; t[4] = r13;
                    r10 = t;
                    r11 = r3;
                    r12 = r4;
                    t = alloc_no_clear ((TC_USEROBJ+0), 4);
                    t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                    r9 = t;
                    PXLL_RETURN(9);
                  } break;
                  default: {
                    goto fatbar_6;
                  }
                }
                PXLL_RETURN(5);
                goto fatbar_6_over;
                fatbar_6:
                r5 = r2;
                switch (get_case_noint (r5)) {
                  case (TC_USEROBJ+0): {
                    r5 = r2;
                    r5 = UOBJ_GET(r5,0);
                    r6 = r2;
                    r6 = UOBJ_GET(r6,1);
                    r7 = r2;
                    r7 = UOBJ_GET(r7,2);
                    r8 = r2;
                    r8 = UOBJ_GET(r8,3);
                    r9 = r1;
                    r10 = r5;
                    r11 = r3;
                    r12 = r4;
                    t = alloc_no_clear ((TC_USEROBJ+4), 4);
                    t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                    r9 = t;
                    r10 = r6;
                    r11 = varref (0,1);
                    r12 = varref (0,2);
                    r13 = varref (0,3);
                    t = alloc_no_clear ((TC_USEROBJ+4), 4);
                    t[1] = r10; t[2] = r11; t[3] = r12; t[4] = r13;
                    r10 = t;
                    r11 = r7;
                    r12 = r8;
                    t = alloc_no_clear ((TC_USEROBJ+0), 4);
                    t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                    r9 = t;
                    PXLL_RETURN(9);
                  } break;
                  default: {
                    goto fatbar_5;
                  }
                }
                PXLL_RETURN(5);
                fatbar_6_over: ;
                PXLL_RETURN(5);
              } break;
              default: {
                goto fatbar_5;
              }
            }
            PXLL_RETURN(1);
            goto fatbar_5_over;
            fatbar_5:
            r1 = r0;
            r2 = varref (0,1);
            r3 = varref (0,2);
            r4 = varref (0,3);
            t = alloc_no_clear ((TC_USEROBJ+4), 4);
            t[1] = r1; t[2] = r2; t[3] = r3; t[4] = r4;
            r1 = t;
            PXLL_RETURN(1);
            fatbar_5_over: ;
            PXLL_RETURN(1);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = varref (1,2);
            r1[2] = r2;
            r2 = varref (4,1);
            r1[3] = r2;
            r2 = top[13];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L26; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
            L26:
            r0 = k[4];; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = varref (0,0);
            r0 = PXLL_TEST(unbox(r0)<0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r1 = varref (0,1);
              r0[2] = r1;
              r1 = varref (2,0);
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L27;  k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_ins_439_i52;
              L27:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              r1 = r0;
              switch (get_case_noint (r1)) {
                case (TC_USEROBJ+0): {
                  r1 = r0;
                  r1 = UOBJ_GET(r1,0);
                  r2 = r0;
                  r2 = UOBJ_GET(r2,1);
                  r3 = r0;
                  r3 = UOBJ_GET(r3,2);
                  r4 = r0;
                  r4 = UOBJ_GET(r4,3);
                  r5 = r1;
                  switch (get_case_noint (r5)) {
                    case (TC_USEROBJ+0): {
                      r5 = r1;
                      r5 = UOBJ_GET(r5,0);
                      r6 = r1;
                      r6 = UOBJ_GET(r6,1);
                      r7 = r1;
                      r7 = UOBJ_GET(r7,2);
                      r8 = r1;
                      r8 = UOBJ_GET(r8,3);
                      r9 = varref (0,0);
                      r10 = r5;
                      r11 = varref (0,2);
                      r12 = varref (0,3);
                      t = alloc_no_clear ((TC_USEROBJ+4), 4);
                      t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                      r9 = t;
                      r10 = r6;
                      r11 = r2;
                      r12 = r3;
                      r13 = r4;
                      t = alloc_no_clear ((TC_USEROBJ+4), 4);
                      t[1] = r10; t[2] = r11; t[3] = r12; t[4] = r13;
                      r10 = t;
                      r11 = r7;
                      r12 = r8;
                      t = alloc_no_clear ((TC_USEROBJ+0), 4);
                      t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                      r9 = t;
                      PXLL_RETURN(9);
                    } break;
                    default: {
                      goto fatbar_8;
                    }
                  }
                  PXLL_RETURN(5);
                  goto fatbar_8_over;
                  fatbar_8:
                  r5 = r2;
                  switch (get_case_noint (r5)) {
                    case (TC_USEROBJ+0): {
                      r5 = r2;
                      r5 = UOBJ_GET(r5,0);
                      r6 = r2;
                      r6 = UOBJ_GET(r6,1);
                      r7 = r2;
                      r7 = UOBJ_GET(r7,2);
                      r8 = r2;
                      r8 = UOBJ_GET(r8,3);
                      r9 = varref (0,0);
                      r10 = r1;
                      r11 = varref (0,2);
                      r12 = varref (0,3);
                      t = alloc_no_clear ((TC_USEROBJ+4), 4);
                      t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                      r9 = t;
                      r10 = r5;
                      r11 = r6;
                      r12 = r7;
                      r13 = r8;
                      t = alloc_no_clear ((TC_USEROBJ+4), 4);
                      t[1] = r10; t[2] = r11; t[3] = r12; t[4] = r13;
                      r10 = t;
                      r11 = r3;
                      r12 = r4;
                      t = alloc_no_clear ((TC_USEROBJ+0), 4);
                      t[1] = r9; t[2] = r10; t[3] = r11; t[4] = r12;
                      r9 = t;
                      PXLL_RETURN(9);
                    } break;
                    default: {
                      goto fatbar_7;
                    }
                  }
                  PXLL_RETURN(5);
                  fatbar_8_over: ;
                  PXLL_RETURN(5);
                } break;
                default: {
                  goto fatbar_7;
                }
              }
              PXLL_RETURN(1);
              goto fatbar_7_over;
              fatbar_7:
              r1 = varref (0,0);
              r2 = r0;
              r3 = varref (0,2);
              r4 = varref (0,3);
              t = alloc_no_clear ((TC_USEROBJ+4), 4);
              t[1] = r1; t[2] = r2; t[3] = r3; t[4] = r4;
              r1 = t;
              PXLL_RETURN(1);
              fatbar_7_over: ;
              PXLL_RETURN(1);
            } else {
              r0 = varref (1,0);
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
        case (TC_USEROBJ+0): {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r1 = varref (1,0);
          r1 = UOBJ_GET(r1,3);
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = varref (4,1);
          r1[2] = r2;
          r2 = varref (1,2);
          r1[3] = r2;
          r2 = top[13];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L28; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
          L28:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = varref (0,0);
          r0 = PXLL_TEST(unbox(r0)<0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (0,0);
            r0[2] = r1;
            r1 = varref (2,0);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L29;  k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_ins_439_i52;
            L29:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            r1 = varref (0,1);
            r2 = varref (0,2);
            r3 = varref (0,3);
            t = alloc_no_clear ((TC_USEROBJ+0), 4);
            t[1] = r0; t[2] = r1; t[3] = r2; t[4] = r3;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = varref (1,2);
            r1[2] = r2;
            r2 = varref (4,1);
            r1[3] = r2;
            r2 = top[13];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L30; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
            L30:
            r0 = k[4];; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = varref (0,0);
            r0 = PXLL_TEST(unbox(r0)<0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r1 = varref (0,1);
              r0[2] = r1;
              r1 = varref (2,0);
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L31;  k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_ins_439_i52;
              L31:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              r1 = varref (0,0);
              r2 = varref (0,2);
              r3 = varref (0,3);
              t = alloc_no_clear ((TC_USEROBJ+0), 4);
              t[1] = r1; t[2] = r0; t[3] = r2; t[4] = r3;
              r0 = t;
              PXLL_RETURN(0);
            } else {
              r0 = varref (1,0);
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = (object*)UITAG(2);
          r1 = (object*)UITAG(2);
          r2 = varref (2,1);
          r3 = varref (3,0);
          t = alloc_no_clear ((TC_USEROBJ+0), 4);
          t[1] = r0; t[2] = r1; t[3] = r2; t[4] = r3;
          r0 = t;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L23:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_ins_439_i52; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (2,0);
    r1[2] = r2;
    r2 = varref (1,0);
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L32; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_ins_439_i52;
    L32:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = varref (0,0);
    switch (get_case_noint (r0)) {
      case (TC_USEROBJ+0): {
        r0 = varref (0,0);
        r0 = UOBJ_GET(r0,0);
        r1 = varref (0,0);
        r1 = UOBJ_GET(r1,1);
        r2 = varref (0,0);
        r2 = UOBJ_GET(r2,2);
        r3 = varref (0,0);
        r3 = UOBJ_GET(r3,3);
        r4 = r0;
        r5 = r1;
        r6 = r2;
        r7 = r3;
        t = alloc_no_clear ((TC_USEROBJ+4), 4);
        t[1] = r4; t[2] = r5; t[3] = r6; t[4] = r7;
        r0 = t;
      } break;
      default: {
        goto fatbar_4;
      }
    }
    goto fatbar_4_over;
    fatbar_4:
    r0 = varref (0,0);
    fatbar_4_over: ;
    lenv = lenv[1];
    lenv = lenv[1];
    lenv = lenv[1];
    top[10] = r0;
    r0 = varref (0,0);
    PXLL_RETURN(0);
  L22:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_intern_symbol_127; r1[2] = lenv;
  r0[14] = r1;
  r1 = box(O_RDONLY);
  r0[15] = r1;
  // def 'reader_132'
  goto L33;
  FUN_reader_132:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) 65794;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 24);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object*) constructed_75[0];
    r1[2] = r2;
    r2 = top[2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L34; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_char_class_131;
    L34:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    // def 'peek_548'
    goto L35;
    FUN_peek_548:
      check_heap (0);
      r0 = varref (1,0);
      r1 = r0;
      r2 = (object *) 65794;
      r0 = PXLL_TEST(r1==r2);
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (2,0);
        r0[2] = r1;
        r1 = top[4];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L36;  k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
        L36:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        varset (1, 0, r0);
      } else {
        r0 = (object *) 14;
      }
      r0 = varref (1,0);
      PXLL_RETURN(0);
    L35:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_peek_548; r1[2] = lenv;
    r0[3] = r1;
    r1 = (object*) constructed_74[0];
    r0[4] = r1;
    // def 'read-hex-digit_557'
    goto L37;
    FUN_read_hex_digit_557:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = varref (2,2);
      r1[2] = r2;
      r2 = varref (1,0);
      r1[3] = r2;
      r2 = top[6];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L38; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_118;
      L38:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (0,0);
      switch (get_case_noint (r0)) {
        case (TC_USEROBJ+0): {
          r0 = varref (0,0);
          r0 = UOBJ_GET(r0,0);
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = (object*) &constructed_73;
          dump_object (r0, 0); fprintf (stdout, "\n");
          goto Lreturn;
          r0 = PXLL_UNDEFINED;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L37:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_digit_557; r1[2] = lenv;
    r0[5] = r1;
    // def 'read-hex-code_558'
    goto L39;
    FUN_read_hex_code_558:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = varref (3,0);
      r2[2] = r3;
      r2 = allocate (TC_TUPLE, 2);
      r3 = varref (4,0);
      r2[2] = r3;
      r3 = top[4];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L40; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_file_2fread_char_115;
      L40:
      r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
      r2 = result;
      varset (3, 0, r2);
      r2 = varref (0,0);
      lenv = lenv[1];
      r1[2] = r2;
      r2 = varref (1,3);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L41; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_read_hex_digit_557;
      L41:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = varref (3,0);
      r2[2] = r3;
      r2 = allocate (TC_TUPLE, 2);
      r3 = varref (4,0);
      r2[2] = r3;
      r3 = top[4];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L42; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_file_2fread_char_115;
      L42:
      r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
      r2 = result;
      varset (3, 0, r2);
      r2 = varref (0,0);
      lenv = lenv[1];
      r1[2] = r2;
      r2 = varref (1,3);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L43; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_read_hex_digit_557;
      L43:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = varref (0,0);
      r2 = (object *) 17;
      r1 = box(unbox(r1)<<unbox(r2));
      r2 = r1;
      r3 = varref (0,1);
      r1 = box(unbox(r2)+unbox(r3));
      r0[4] = r1;
      r0 = varref (0,2);
      r0 = TO_CHAR(unbox(r0));
      PXLL_RETURN(0);
    L39:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_code_558; r1[2] = lenv;
    r0[6] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object*) &constructed_72;
    r1[2] = r2;
    r2 = top[9];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L44; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_98;
    L44:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[7] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object*) &constructed_71;
    r1[2] = r2;
    r2 = top[9];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L45; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_98;
    L45:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[8] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (0,6);
    r1[2] = r2;
    r2 = top[2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L46; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_char_class_131;
    L46:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[9] = r1;
    r1 = (object*) constructed_70[0];
    r0[10] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = varref (0,8);
    r2[2] = r3;
    r3 = (object *) TC_NIL;
    r2[3] = r3;
    r3 = top[5];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L47; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_69;
    L47:
    r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = varref (0,5);
    r1[3] = r2;
    r2 = top[5];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L48; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
    L48:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[11] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = varref (0,9);
    r1[2] = r2;
    r2 = top[2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L49; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_char_class_131;
    L49:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[12] = r1;
    r1 = (object*) constructed_69[0];
    r0[13] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object*) &constructed_68;
    r1[2] = r2;
    r2 = top[9];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L50; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_98;
    L50:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[14] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = allocate (TC_TUPLE, 3);
    r4 = allocate (TC_TUPLE, 3);
    r5 = varref (1,12);
    r4[2] = r5;
    r5 = (object *) TC_NIL;
    r4[3] = r5;
    r5 = top[5];
    t = allocate (TC_SAVE, 7);
    t[1] = k; t[2] = lenv; t[3] = &&L51; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
    r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_69;
    L51:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7];; lenv = k[2]; k = k[1];
    r4 = result;
    r3[2] = r4;
    r4 = varref (1,6);
    r3[3] = r4;
    r4 = top[5];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L52; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_69;
    L52:
    r0 = k[4]; r1 = k[5]; r2 = k[6];; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r2 = (object *) 11522;
    r3 = varref (0,0);
    t = alloc_no_clear (TC_PAIR, 2);
    t[1] = r2; t[2] = r3;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L53; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_char_class_131;
    L53:
    r0 = k[4];; lenv = k[2]; k = k[1];
    r1 = result;
    r0[15] = r1;
    // def 'skip-whitespace_552'
    goto L54;
    FUN_skip_whitespace_552:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def 'loop_581'
      goto L55;
      FUN_loop_581:
        check_heap (0);
        r0 = varref (0,0);
        r1 = (object *) 65794;
        if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = varref (0,0);
          r1 = (object *) 15106;
          if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def 'loop_579_i78'
            goto L56;
            FUN_loop_579_i78:
              check_heap (0);
              r0 = (object *) 3330;
              r1 = varref (0,0);
              if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = (object *) 2562;
                r1 = varref (0,0);
                if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
                  r0 = (object *) 14;
                  PXLL_RETURN(0);
                } else {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = varref (6,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = varref (7,0);
                  r0[2] = r1;
                  r1 = top[4];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L57;  k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                  L57:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  varset (6, 0, r0);
                  r0 = varref (0,0);
                  lenv = lenv[1];
                  r0 = varref (4,1);
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L58;  k = t;
                  lenv = r0[2]; goto FUN_peek_548;
                  L58:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  lenv[2] = r0;
                  goto FUN_loop_579_i78;
                }
                PXLL_RETURN(0);
              }
              PXLL_RETURN(0);
            L56:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_loop_579_i78; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = varref (5,0);
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = varref (6,0);
            r1[2] = r2;
            r2 = top[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L59; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_file_2fread_char_115;
            L59:
            r0 = k[4];; lenv = k[2]; k = k[1];
            r1 = result;
            varset (5, 0, r1);
            r1 = varref (0,0);
            lenv = lenv[1];
            r0[2] = r1;
            r1 = varref (0,0);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L60;  k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_loop_579_i78;
            L60:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            r0 = varref (2,1);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L61;  k = t;
            lenv = r0[2]; goto FUN_peek_548;
            L61:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv[2] = r0;
            goto FUN_loop_581;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (0,0);
            r0[2] = r1;
            r1 = varref (2,0);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L62;  k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L62:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = varref (4,0);
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r1 = varref (5,0);
              r0[2] = r1;
              r1 = top[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L63;  k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
              L63:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              varset (4, 0, r0);
              r0 = varref (0,0);
              lenv = lenv[1];
              r0 = varref (2,1);
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L64;  k = t;
              lenv = r0[2]; goto FUN_peek_548;
              L64:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              lenv[2] = r0;
              goto FUN_loop_581;
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L55:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_581; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (1,1);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L65; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_peek_548;
      L65:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = varref (0,0);
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_581;
    L54:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_skip_whitespace_552; r1[2] = lenv;
    r0[16] = r1;
    // def 'read-atom_553'
    goto L66;
    FUN_read_atom_553:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) PXLL_TRUE;
      r0[2] = r1;
      r1 = (object *) PXLL_FALSE;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def 'loop_585'
      goto L67;
      FUN_loop_585:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (4,1);
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L68; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_peek_548;
        L68:
        r0 = k[4];; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = varref (3,1);
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (0,0);
          r0[2] = r1;
          r1 = varref (4,10);
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L69;  k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L69:
          ; lenv = k[2]; k = k[1];
          r0 = result;
        } else {
          r0 = (object *) PXLL_FALSE;
        }
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r2 = varref (1,0);
          r1[2] = r2;
          r2 = (object *) TC_NIL;
          r1[3] = r2;
          r2 = top[5];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L70; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
          L70:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = top[8];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L71;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_97;
          L71:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          r1 = allocate (TC_TUPLE, 2);
          r2 = varref (1,0);
          r1[2] = r2;
          r2 = top[7];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L72; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_72;
          L72:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r2 = varref (3,0);
          t = alloc_no_clear ((TC_USEROBJ+0), 3);
          t[1] = r0; t[2] = r2; t[3] = r1;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) PXLL_TRUE;
          varset (3, 1, r0);
          r0 = varref (3,0);
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (0,0);
            r0[2] = r1;
            r1 = varref (4,7);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L73;  k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L73:
            ; lenv = k[2]; k = k[1];
            r0 = result;
          } else {
            r0 = (object *) PXLL_FALSE;
          }
          varset (3, 0, r0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (6,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (7,0);
          r0[2] = r1;
          r1 = top[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L74;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
          L74:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          varset (6, 0, r0);
          r0 = varref (0,0);
          lenv = lenv[1];
          r1 = varref (1,0);
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r0; t[2] = r1;
          r0 = t;
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_loop_585;
        }
        PXLL_RETURN(0);
      L67:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_585; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) constructed_67[0];
      r0[2] = r1;
      r1 = varref (0,0);
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_585;
    L66:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_atom_553; r1[2] = lenv;
    r0[17] = r1;
    // def 'read1_554'
    goto L75;
    FUN_read1_554:
      check_heap (0);
      r0 = varref (0,14);
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L76;  k = t;
      lenv = r0[2]; goto FUN_skip_whitespace_552;
      L76:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (1,1);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L77; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_peek_548;
      L77:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (0,0);
      r1 = (object *) 65794;
      r2 = r0;
      r0 = PXLL_TEST(r1==r2);
      if PXLL_IS_TRUE(r0) {
        r0 = (object*) &constructed_52;
        dump_object (r0, 0); fprintf (stdout, "\n");
        goto Lreturn;
        r0 = PXLL_UNDEFINED;
        PXLL_RETURN(0);
      } else {
        r0 = varref (0,0);
        r1 = (object *) 10242;
        r2 = r0;
        r0 = PXLL_TEST(r1==r2);
        if PXLL_IS_TRUE(r0) {
          r0 = varref (1,19);
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L78;  k = t;
          lenv = r0[2]; goto FUN_read_list_560;
          L78:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          t = alloc_no_clear ((TC_USEROBJ+0), 1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = varref (0,0);
          r1 = (object *) 31490;
          r2 = r0;
          r0 = PXLL_TEST(r1==r2);
          if PXLL_IS_TRUE(r0) {
            r0 = varref (1,20);
            lenv = r0[2]; goto FUN_read_record_561;
          } else {
            r0 = varref (0,0);
            r1 = (object *) 8706;
            r2 = r0;
            r0 = PXLL_TEST(r1==r2);
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = varref (3,0);
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r1 = varref (4,0);
              r0[2] = r1;
              r1 = top[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L79;  k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
              L79:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              varset (3, 0, r0);
              r0 = varref (0,0);
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              // def 'loop_609_i87'
              goto L80;
              FUN_loop_609_i87:
                check_heap (0);
                r0 = varref (0,0);
                r1 = (object *) 8706;
                r2 = r0;
                r0 = PXLL_TEST(r1==r2);
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = varref (5,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = varref (6,0);
                  r0[2] = r1;
                  r1 = top[4];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L81;  k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                  L81:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  varset (5, 0, r0);
                  r0 = varref (0,0);
                  lenv = lenv[1];
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = varref (0,1);
                  r1[2] = r2;
                  r2 = (object *) TC_NIL;
                  r1[3] = r2;
                  r2 = top[5];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L82; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
                  L82:
                  r0 = k[4];; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = top[8];
                  r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_97;
                } else {
                  r0 = varref (0,0);
                  r1 = (object *) 23554;
                  r2 = r0;
                  r0 = PXLL_TEST(r1==r2);
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (5,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = varref (6,0);
                    r0[2] = r1;
                    r1 = top[4];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L83;  k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                    L83:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    varset (5, 0, r0);
                    r0 = varref (0,0);
                    lenv = lenv[1];
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (5,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = varref (6,0);
                    r0[2] = r1;
                    r1 = top[4];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L84;  k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                    L84:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    varset (5, 0, r0);
                    r0 = varref (0,0);
                    lenv = lenv[1];
                    varset (0, 0, r0);
                    r0 = varref (0,0);
                    r1 = (object *) 30722;
                    r2 = r0;
                    r0 = PXLL_TEST(r1==r2);
                    if PXLL_IS_TRUE(r0) {
                      r0 = varref (3,4);
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L85;  k = t;
                      lenv = r0[2]; goto FUN_read_hex_code_558;
                      L85:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      r1 = varref (0,1);
                      t = alloc_no_clear (TC_PAIR, 2);
                      t[1] = r0; t[2] = r1;
                      r0 = t;
                      r1 = varref (3,1);
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L86; t[4] = r0; k = t;
                      lenv = r1[2]; goto FUN_peek_548;
                      L86:
                      r0 = k[4];; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv[2] = r1;
                      lenv[3] = r0;
                      goto FUN_loop_609_i87;
                    } else {
                      r0 = varref (0,0);
                      r1 = (object *) 22530;
                      r2 = r0;
                      r0 = PXLL_TEST(r1==r2);
                      if PXLL_IS_TRUE(r0) {
                        r0 = varref (3,4);
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L87;  k = t;
                        lenv = r0[2]; goto FUN_read_hex_code_558;
                        L87:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        r1 = varref (0,1);
                        t = alloc_no_clear (TC_PAIR, 2);
                        t[1] = r0; t[2] = r1;
                        r0 = t;
                        r1 = varref (3,1);
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L88; t[4] = r0; k = t;
                        lenv = r1[2]; goto FUN_peek_548;
                        L88:
                        r0 = k[4];; lenv = k[2]; k = k[1];
                        r1 = result;
                        lenv[2] = r1;
                        lenv[3] = r0;
                        goto FUN_loop_609_i87;
                      } else {
                        r0 = varref (0,0);
                        r1 = (object *) 28162;
                        r2 = r0;
                        r0 = PXLL_TEST(r1==r2);
                        if PXLL_IS_TRUE(r0) {
                          r0 = (object *) 2562;
                          r1 = varref (0,1);
                          t = alloc_no_clear (TC_PAIR, 2);
                          t[1] = r0; t[2] = r1;
                          r0 = t;
                          r1 = varref (3,1);
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L89; t[4] = r0; k = t;
                          lenv = r1[2]; goto FUN_peek_548;
                          L89:
                          r0 = k[4];; lenv = k[2]; k = k[1];
                          r1 = result;
                          lenv[2] = r1;
                          lenv[3] = r0;
                          goto FUN_loop_609_i87;
                        } else {
                          r0 = varref (0,0);
                          r1 = (object *) 29698;
                          r2 = r0;
                          r0 = PXLL_TEST(r1==r2);
                          if PXLL_IS_TRUE(r0) {
                            r0 = (object *) 2306;
                            r1 = varref (0,1);
                            t = alloc_no_clear (TC_PAIR, 2);
                            t[1] = r0; t[2] = r1;
                            r0 = t;
                            r1 = varref (3,1);
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L90; t[4] = r0; k = t;
                            lenv = r1[2]; goto FUN_peek_548;
                            L90:
                            r0 = k[4];; lenv = k[2]; k = k[1];
                            r1 = result;
                            lenv[2] = r1;
                            lenv[3] = r0;
                            goto FUN_loop_609_i87;
                          } else {
                            r0 = varref (0,0);
                            r1 = (object *) 8706;
                            r2 = r0;
                            r0 = PXLL_TEST(r1==r2);
                            if PXLL_IS_TRUE(r0) {
                              r0 = (object *) 8706;
                              r1 = varref (0,1);
                              t = alloc_no_clear (TC_PAIR, 2);
                              t[1] = r0; t[2] = r1;
                              r0 = t;
                              r1 = varref (3,1);
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L91; t[4] = r0; k = t;
                              lenv = r1[2]; goto FUN_peek_548;
                              L91:
                              r0 = k[4];; lenv = k[2]; k = k[1];
                              r1 = result;
                              lenv[2] = r1;
                              lenv[3] = r0;
                              goto FUN_loop_609_i87;
                            } else {
                              r0 = varref (0,0);
                              r1 = (object *) 23554;
                              r2 = r0;
                              r0 = PXLL_TEST(r1==r2);
                              if PXLL_IS_TRUE(r0) {
                                r0 = (object *) 23554;
                                r1 = varref (0,1);
                                t = alloc_no_clear (TC_PAIR, 2);
                                t[1] = r0; t[2] = r1;
                                r0 = t;
                                r1 = varref (3,1);
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L92; t[4] = r0; k = t;
                                lenv = r1[2]; goto FUN_peek_548;
                                L92:
                                r0 = k[4];; lenv = k[2]; k = k[1];
                                r1 = result;
                                lenv[2] = r1;
                                lenv[3] = r0;
                                goto FUN_loop_609_i87;
                              } else {
                                r0 = (object*) &constructed_54;
                                dump_object (r0, 0); fprintf (stdout, "\n");
                                goto Lreturn;
                                r0 = PXLL_UNDEFINED;
                                PXLL_RETURN(0);
                              }
                              PXLL_RETURN(0);
                            }
                            PXLL_RETURN(0);
                          }
                          PXLL_RETURN(0);
                        }
                        PXLL_RETURN(0);
                      }
                      PXLL_RETURN(0);
                    }
                    PXLL_RETURN(0);
                  } else {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (5,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = varref (6,0);
                    r0[2] = r1;
                    r1 = top[4];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L93;  k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                    L93:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    varset (5, 0, r0);
                    r0 = varref (0,0);
                    lenv = lenv[1];
                    r0 = varref (3,1);
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L94;  k = t;
                    lenv = r0[2]; goto FUN_peek_548;
                    L94:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    r1 = varref (0,0);
                    r2 = varref (0,1);
                    t = alloc_no_clear (TC_PAIR, 2);
                    t[1] = r1; t[2] = r2;
                    r1 = t;
                    lenv[2] = r0;
                    lenv[3] = r1;
                    goto FUN_loop_609_i87;
                  }
                  PXLL_RETURN(0);
                }
                PXLL_RETURN(0);
              L80:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_loop_609_i87; r1[2] = lenv;
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r1 = varref (2,1);
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L95; t[4] = r0; k = t;
              lenv = r1[2]; goto FUN_peek_548;
              L95:
              r0 = k[4];; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = (object*) constructed_53[0];
              r0[3] = r1;
              r1 = varref (0,0);
              r0[1] = r1[2]; lenv = r0; goto FUN_loop_609_i87;
            } else {
              r0 = varref (0,0);
              r1 = (object *) 9986;
              r2 = r0;
              r0 = PXLL_TEST(r1==r2);
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = varref (3,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r1 = varref (4,0);
                r0[2] = r1;
                r1 = top[4];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L96;  k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                L96:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                varset (3, 0, r0);
                r0 = varref (0,0);
                lenv = lenv[1];
                r0 = varref (1,17);
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L97;  k = t;
                lenv = r0[2]; goto FUN_read_555;
                L97:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                r1 = (object *) TC_NIL;
                t = alloc_no_clear (TC_PAIR, 2);
                t[1] = r0; t[2] = r1;
                r0 = t;
                r1 = (object*) constructed_28[0];
                t = alloc_no_clear ((TC_USEROBJ+4), 1);
                t[1] = r1;
                r1 = t;
                t = alloc_no_clear (TC_PAIR, 2);
                t[1] = r1; t[2] = r0;
                r0 = t;
                t = alloc_no_clear ((TC_USEROBJ+0), 1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } else {
                r0 = varref (0,0);
                r1 = (object *) 11266;
                r2 = r0;
                r0 = PXLL_TEST(r1==r2);
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = varref (3,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = varref (4,0);
                  r0[2] = r1;
                  r1 = top[4];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L98;  k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                  L98:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  varset (3, 0, r0);
                  r0 = varref (0,0);
                  lenv = lenv[1];
                  r0 = varref (1,17);
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L99;  k = t;
                  lenv = r0[2]; goto FUN_read_555;
                  L99:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  r1 = (object *) TC_NIL;
                  t = alloc_no_clear (TC_PAIR, 2);
                  t[1] = r0; t[2] = r1;
                  r0 = t;
                  r1 = (object*) constructed_56[0];
                  t = alloc_no_clear ((TC_USEROBJ+4), 1);
                  t[1] = r1;
                  r1 = t;
                  t = alloc_no_clear (TC_PAIR, 2);
                  t[1] = r1; t[2] = r0;
                  r0 = t;
                  t = alloc_no_clear ((TC_USEROBJ+0), 1);
                  t[1] = r0;
                  r0 = t;
                  PXLL_RETURN(0);
                } else {
                  r0 = varref (0,0);
                  r1 = (object *) 14850;
                  r2 = r0;
                  r0 = PXLL_TEST(r1==r2);
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (3,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = varref (4,0);
                    r0[2] = r1;
                    r1 = top[4];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L100;  k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                    L100:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    varset (3, 0, r0);
                    r0 = varref (0,0);
                    lenv = lenv[1];
                    r0 = varref (1,17);
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L101;  k = t;
                    lenv = r0[2]; goto FUN_read_555;
                    L101:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    r1 = (object *) TC_NIL;
                    t = alloc_no_clear (TC_PAIR, 2);
                    t[1] = r0; t[2] = r1;
                    r0 = t;
                    r1 = (object*) constructed_22[0];
                    t = alloc_no_clear ((TC_USEROBJ+4), 1);
                    t[1] = r1;
                    r1 = t;
                    t = alloc_no_clear (TC_PAIR, 2);
                    t[1] = r1; t[2] = r0;
                    r0 = t;
                    t = alloc_no_clear ((TC_USEROBJ+0), 1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } else {
                    r0 = varref (0,0);
                    r1 = (object *) 8962;
                    r2 = r0;
                    r0 = PXLL_TEST(r1==r2);
                    if PXLL_IS_TRUE(r0) {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = varref (3,0);
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = varref (4,0);
                      r0[2] = r1;
                      r1 = top[4];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L102;  k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                      L102:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      varset (3, 0, r0);
                      r0 = varref (0,0);
                      lenv = lenv[1];
                      r0 = varref (1,1);
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L103;  k = t;
                      lenv = r0[2]; goto FUN_peek_548;
                      L103:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      varset (0, 0, r0);
                      r0 = varref (0,0);
                      r1 = (object *) 23554;
                      r2 = r0;
                      r0 = PXLL_TEST(r1==r2);
                      if PXLL_IS_TRUE(r0) {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = varref (3,0);
                        r0[2] = r1;
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = varref (4,0);
                        r0[2] = r1;
                        r1 = top[4];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L104;  k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
                        L104:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        varset (3, 0, r0);
                        r0 = varref (0,0);
                        lenv = lenv[1];
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = varref (2,15);
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L105; t[4] = r0; k = t;
                        lenv = r1[2]; goto FUN_read_atom_553;
                        L105:
                        r0 = k[4];; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r0 = varref (0,0);
                        switch (get_case_noint (r0)) {
                          case (TC_USEROBJ+0): {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = varref (1,0);
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = varref (1,0);
                            r1 = UOBJ_GET(r1,2);
                            r0[3] = r1;
                            r0 = (object *) 3;
                            r1 = varref (0,1);
                            if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
                              r0 = varref (0,0);
                              r1 = (object *) 1;
                              r0 = TO_CHAR(((unsigned char *)((pxll_string*)(r0))->data)[unbox(r1)]);
                              PXLL_RETURN(0);
                            } else {
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = (object*) &constructed_62;
                              r1[2] = r2;
                              r2 = varref (1,0);
                              r1[3] = r2;
                              r2 = top[13];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L106; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                              L106:
                              r0 = k[4];; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[2] = r1;
                              r0 = varref (0,0);
                              r0 = PXLL_TEST(unbox(r0)==0);
                              lenv = lenv[1];
                              if PXLL_IS_TRUE(r0) {
                                r0 = (object *) 2562;
                                PXLL_RETURN(0);
                              } else {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = allocate (TC_TUPLE, 3);
                                r2 = (object*) &constructed_61;
                                r1[2] = r2;
                                r2 = varref (1,0);
                                r1[3] = r2;
                                r2 = top[13];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L107; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                L107:
                                r0 = k[4];; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[2] = r1;
                                r0 = varref (0,0);
                                r0 = PXLL_TEST(unbox(r0)==0);
                                lenv = lenv[1];
                                if PXLL_IS_TRUE(r0) {
                                  r0 = (object *) 8194;
                                  PXLL_RETURN(0);
                                } else {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = (object*) &constructed_60;
                                  r1[2] = r2;
                                  r2 = varref (1,0);
                                  r1[3] = r2;
                                  r2 = top[13];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L108; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                  L108:
                                  r0 = k[4];; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r0 = varref (0,0);
                                  r0 = PXLL_TEST(unbox(r0)==0);
                                  lenv = lenv[1];
                                  if PXLL_IS_TRUE(r0) {
                                    r0 = (object *) 3330;
                                    PXLL_RETURN(0);
                                  } else {
                                    r0 = allocate (TC_TUPLE, 2);
                                    r0[1] = lenv; lenv = r0;
                                    r1 = allocate (TC_TUPLE, 3);
                                    r2 = (object*) &constructed_59;
                                    r1[2] = r2;
                                    r2 = varref (1,0);
                                    r1[3] = r2;
                                    r2 = top[13];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L109; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                    L109:
                                    r0 = k[4];; lenv = k[2]; k = k[1];
                                    r1 = result;
                                    r0[2] = r1;
                                    r0 = varref (0,0);
                                    r0 = PXLL_TEST(unbox(r0)==0);
                                    lenv = lenv[1];
                                    if PXLL_IS_TRUE(r0) {
                                      r0 = (object *) 2306;
                                      PXLL_RETURN(0);
                                    } else {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = (object*) &constructed_58;
                                      r1[2] = r2;
                                      r2 = varref (1,0);
                                      r1[3] = r2;
                                      r2 = top[13];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L110; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                      L110:
                                      r0 = k[4];; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r0 = varref (0,0);
                                      r0 = PXLL_TEST(unbox(r0)==0);
                                      lenv = lenv[1];
                                      if PXLL_IS_TRUE(r0) {
                                        r0 = (object *) 65794;
                                        PXLL_RETURN(0);
                                      } else {
                                        r0 = allocate (TC_TUPLE, 2);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = allocate (TC_TUPLE, 3);
                                        r2 = (object*) &constructed_57;
                                        r1[2] = r2;
                                        r2 = varref (1,0);
                                        r1[3] = r2;
                                        r2 = top[13];
                                        t = allocate (TC_SAVE, 4);
                                        t[1] = k; t[2] = lenv; t[3] = &&L111; t[4] = r0; k = t;
                                        r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                        L111:
                                        r0 = k[4];; lenv = k[2]; k = k[1];
                                        r1 = result;
                                        r0[2] = r1;
                                        r0 = varref (0,0);
                                        r0 = PXLL_TEST(unbox(r0)==0);
                                        lenv = lenv[1];
                                        if PXLL_IS_TRUE(r0) {
                                          r0 = (object *) 2;
                                          PXLL_RETURN(0);
                                        } else {
                                          lenv = ((object *)lenv)[1];
                                          goto fatbar_3;
                                        }
                                        PXLL_RETURN(0);
                                      }
                                      PXLL_RETURN(0);
                                    }
                                    PXLL_RETURN(0);
                                  }
                                  PXLL_RETURN(0);
                                }
                                PXLL_RETURN(0);
                              }
                              PXLL_RETURN(0);
                            }
                            PXLL_RETURN(0);
                          } break;
                        }
                        PXLL_RETURN(0);
                        goto fatbar_3_over;
                        fatbar_3:
                        r0 = (object*) &constructed_63;
                        dump_object (r0, 0); fprintf (stdout, "\n");
                        goto Lreturn;
                        r0 = PXLL_UNDEFINED;
                        PXLL_RETURN(0);
                        fatbar_3_over: ;
                        PXLL_RETURN(0);
                      } else {
                        r0 = varref (0,0);
                        r1 = (object *) 21506;
                        r2 = r0;
                        r0 = PXLL_TEST(r1==r2);
                        if PXLL_IS_TRUE(r0) {
                          r0 = (object *) PXLL_TRUE;
                          t = alloc_no_clear ((TC_USEROBJ+16), 1);
                          t[1] = r0;
                          r0 = t;
                          PXLL_RETURN(0);
                        } else {
                          r0 = varref (0,0);
                          r1 = (object *) 29698;
                          r2 = r0;
                          r0 = PXLL_TEST(r1==r2);
                          if PXLL_IS_TRUE(r0) {
                            r0 = (object *) PXLL_TRUE;
                            t = alloc_no_clear ((TC_USEROBJ+16), 1);
                            t[1] = r0;
                            r0 = t;
                            PXLL_RETURN(0);
                          } else {
                            r0 = varref (0,0);
                            r1 = (object *) 17922;
                            r2 = r0;
                            r0 = PXLL_TEST(r1==r2);
                            if PXLL_IS_TRUE(r0) {
                              r0 = (object *) PXLL_FALSE;
                              t = alloc_no_clear ((TC_USEROBJ+16), 1);
                              t[1] = r0;
                              r0 = t;
                              PXLL_RETURN(0);
                            } else {
                              r0 = varref (0,0);
                              r1 = (object *) 26114;
                              r2 = r0;
                              r0 = PXLL_TEST(r1==r2);
                              if PXLL_IS_TRUE(r0) {
                                r0 = (object *) PXLL_FALSE;
                                t = alloc_no_clear ((TC_USEROBJ+16), 1);
                                t[1] = r0;
                                r0 = t;
                                PXLL_RETURN(0);
                              } else {
                                r0 = varref (0,0);
                                r1 = (object *) 21762;
                                r2 = r0;
                                r0 = PXLL_TEST(r1==r2);
                                if PXLL_IS_TRUE(r0) {
                                  r0 = (object *) 14;
                                  PXLL_RETURN(0);
                                } else {
                                  r0 = varref (0,0);
                                  r1 = (object *) 29954;
                                  r2 = r0;
                                  r0 = PXLL_TEST(r1==r2);
                                  if PXLL_IS_TRUE(r0) {
                                    r0 = (object *) 14;
                                    PXLL_RETURN(0);
                                  } else {
                                    r0 = varref (0,0);
                                    r1 = (object *) 10242;
                                    r2 = r0;
                                    r0 = PXLL_TEST(r1==r2);
                                    if PXLL_IS_TRUE(r0) {
                                      r0 = varref (1,19);
                                      t = allocate (TC_SAVE, 3);
                                      t[1] = k; t[2] = lenv; t[3] = &&L112;  k = t;
                                      lenv = r0[2]; goto FUN_read_list_560;
                                      L112:
                                      ; lenv = k[2]; k = k[1];
                                      r0 = result;
                                      t = alloc_no_clear ((TC_USEROBJ+28), 1);
                                      t[1] = r0;
                                      r0 = t;
                                      PXLL_RETURN(0);
                                    } else {
                                      r0 = (object*) &constructed_64;
                                      dump_object (r0, 0); fprintf (stdout, "\n");
                                      goto Lreturn;
                                      r0 = PXLL_UNDEFINED;
                                      PXLL_RETURN(0);
                                    }
                                    PXLL_RETURN(0);
                                  }
                                  PXLL_RETURN(0);
                                }
                                PXLL_RETURN(0);
                              }
                              PXLL_RETURN(0);
                            }
                            PXLL_RETURN(0);
                          }
                          PXLL_RETURN(0);
                        }
                        PXLL_RETURN(0);
                      }
                      PXLL_RETURN(0);
                    } else {
                      r0 = varref (0,0);
                      r1 = (object *) 10498;
                      r2 = r0;
                      r0 = PXLL_TEST(r1==r2);
                      if PXLL_IS_TRUE(r0) {
                        r0 = (object*) &constructed_65;
                        dump_object (r0, 0); fprintf (stdout, "\n");
                        goto Lreturn;
                        r0 = PXLL_UNDEFINED;
                        PXLL_RETURN(0);
                      } else {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = varref (2,15);
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L113; t[4] = r0; k = t;
                        lenv = r1[2]; goto FUN_read_atom_553;
                        L113:
                        r0 = k[4];; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r0 = varref (0,0);
                        switch (get_case_noint (r0)) {
                          case (TC_USEROBJ+0): {
                            r0 = allocate (TC_TUPLE, 4);
                            r0[1] = lenv; lenv = r0;
                            r1 = varref (1,0);
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = varref (1,0);
                            r1 = UOBJ_GET(r1,1);
                            r0[3] = r1;
                            r1 = varref (1,0);
                            r1 = UOBJ_GET(r1,2);
                            r0[4] = r1;
                            r0 = varref (0,1);
                            switch ((pxll_int) (r0)) {
                              case (((pxll_int)PXLL_FALSE)): {
                                r0 = allocate (TC_TUPLE, 3);
                                r0[1] = lenv; lenv = r0;
                                r1 = top[10];
                                r0[2] = r1;
                                r1 = allocate (TC_TUPLE, 2);
                                r1[1] = lenv; lenv = r1;
                                // def 'member0_489_i58_i134'
                                goto L114;
                                FUN_member0_489_i58_i134:
                                  check_heap (0);
                                  r0 = varref (0,0);
                                  switch (get_case_noint (r0)) {
                                    case (TC_USEROBJ+4): {
                                      r0 = allocate (TC_TUPLE, 5);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,2);
                                      r0[4] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,3);
                                      r0[5] = r1;
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = varref (5,0);
                                      r1[2] = r2;
                                      r2 = varref (1,2);
                                      r1[3] = r2;
                                      r2 = top[13];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L115; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                      L115:
                                      r0 = k[4];; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r0 = varref (0,0);
                                      r0 = PXLL_TEST(unbox(r0)<0);
                                      lenv = lenv[1];
                                      if PXLL_IS_TRUE(r0) {
                                        r0 = varref (0,0);
                                        lenv = ((object *)lenv)[1];
                                        lenv[2] = r0;
                                        goto FUN_member0_489_i58_i134;
                                      } else {
                                        r0 = allocate (TC_TUPLE, 2);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = allocate (TC_TUPLE, 3);
                                        r2 = varref (1,2);
                                        r1[2] = r2;
                                        r2 = varref (5,0);
                                        r1[3] = r2;
                                        r2 = top[13];
                                        t = allocate (TC_SAVE, 4);
                                        t[1] = k; t[2] = lenv; t[3] = &&L116; t[4] = r0; k = t;
                                        r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                        L116:
                                        r0 = k[4];; lenv = k[2]; k = k[1];
                                        r1 = result;
                                        r0[2] = r1;
                                        r0 = varref (0,0);
                                        r0 = PXLL_TEST(unbox(r0)<0);
                                        lenv = lenv[1];
                                        if PXLL_IS_TRUE(r0) {
                                          r0 = varref (0,1);
                                          lenv = ((object *)lenv)[1];
                                          lenv[2] = r0;
                                          goto FUN_member0_489_i58_i134;
                                        } else {
                                          r0 = varref (0,3);
                                          t = alloc_no_clear ((TC_USEROBJ+0), 1);
                                          t[1] = r0;
                                          r0 = t;
                                          PXLL_RETURN(0);
                                        }
                                        PXLL_RETURN(0);
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    case (TC_USEROBJ+0): {
                                      r0 = allocate (TC_TUPLE, 5);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,2);
                                      r0[4] = r1;
                                      r1 = varref (1,0);
                                      r1 = UOBJ_GET(r1,3);
                                      r0[5] = r1;
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = varref (5,0);
                                      r1[2] = r2;
                                      r2 = varref (1,2);
                                      r1[3] = r2;
                                      r2 = top[13];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L117; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                      L117:
                                      r0 = k[4];; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r0 = varref (0,0);
                                      r0 = PXLL_TEST(unbox(r0)<0);
                                      lenv = lenv[1];
                                      if PXLL_IS_TRUE(r0) {
                                        r0 = varref (0,0);
                                        lenv = ((object *)lenv)[1];
                                        lenv[2] = r0;
                                        goto FUN_member0_489_i58_i134;
                                      } else {
                                        r0 = allocate (TC_TUPLE, 2);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = allocate (TC_TUPLE, 3);
                                        r2 = varref (1,2);
                                        r1[2] = r2;
                                        r2 = varref (5,0);
                                        r1[3] = r2;
                                        r2 = top[13];
                                        t = allocate (TC_SAVE, 4);
                                        t[1] = k; t[2] = lenv; t[3] = &&L118; t[4] = r0; k = t;
                                        r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_92;
                                        L118:
                                        r0 = k[4];; lenv = k[2]; k = k[1];
                                        r1 = result;
                                        r0[2] = r1;
                                        r0 = varref (0,0);
                                        r0 = PXLL_TEST(unbox(r0)<0);
                                        lenv = lenv[1];
                                        if PXLL_IS_TRUE(r0) {
                                          r0 = varref (0,1);
                                          lenv = ((object *)lenv)[1];
                                          lenv[2] = r0;
                                          goto FUN_member0_489_i58_i134;
                                        } else {
                                          r0 = varref (0,3);
                                          t = alloc_no_clear ((TC_USEROBJ+0), 1);
                                          t[1] = r0;
                                          r0 = t;
                                          PXLL_RETURN(0);
                                        }
                                        PXLL_RETURN(0);
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      r0 = (object*)UITAG(1);
                                      PXLL_RETURN(0);
                                    } break;
                                  }
                                  PXLL_RETURN(0);
                                L114:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_member0_489_i58_i134; r2[2] = lenv;
                                r1[2] = r2;
                                r1 = allocate (TC_TUPLE, 2);
                                r2 = varref (1,0);
                                r1[2] = r2;
                                r2 = varref (0,0);
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L119; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_member0_489_i58_i134;
                                L119:
                                r0 = k[4];; lenv = k[2]; k = k[1];
                                r1 = result;
                                lenv = lenv[1];
                                r0[3] = r1;
                                r0 = varref (0,1);
                                switch (get_case_noint (r0)) {
                                  case (TC_USEROBJ+0): {
                                    r0 = varref (0,1);
                                    r0 = UOBJ_GET(r0,0);
                                  } break;
                                  default: {
                                    r0 = allocate (TC_TUPLE, 2);
                                    r1 = varref (1,0);
                                    t = alloc_no_clear (TC_SYMBOL, 1);
                                    t[1] = r1;
                                    r1 = t;
                                    r0[2] = r1;
                                    r1 = top[14];
                                    t = allocate (TC_SAVE, 3);
                                    t[1] = k; t[2] = lenv; t[3] = &&L120;  k = t;
                                    r0[1] = r1[2]; lenv = r0; goto FUN_intern_symbol_127;
                                    L120:
                                    ; lenv = k[2]; k = k[1];
                                    r0 = result;
                                  } break;
                                }
                                lenv = lenv[1];
                                t = alloc_no_clear ((TC_USEROBJ+4), 1);
                                t[1] = r0;
                                r0 = t;
                                PXLL_RETURN(0);
                              } break;
                              default: {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                // def 'loop_626_i135'
                                goto L121;
                                FUN_loop_626_i135:
                                  check_heap (0);
                                  r0 = varref (0,0);
                                  r1 = varref (2,2);
                                  if PXLL_IS_TRUE(PXLL_TEST(unbox(r0)==unbox(r1))) {
                                    r0 = varref (0,1);
                                    PXLL_RETURN(0);
                                  } else {
                                    r0 = allocate (TC_TUPLE, 2);
                                    r0[1] = lenv; lenv = r0;
                                    r1 = allocate (TC_TUPLE, 3);
                                    r2 = varref (6,11);
                                    r1[2] = r2;
                                    r2 = varref (3,0);
                                    r3 = varref (1,0);
                                    r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r3)]);
                                    r1[3] = r2;
                                    r2 = top[6];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L122; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_118;
                                    L122:
                                    r0 = k[4];; lenv = k[2]; k = k[1];
                                    r1 = result;
                                    r0[2] = r1;
                                    r0 = varref (0,0);
                                    switch (get_case_noint (r0)) {
                                      case (TC_USEROBJ+0): {
                                        r0 = varref (0,0);
                                        r0 = UOBJ_GET(r0,0);
                                        r1 = varref (1,1);
                                        r2 = (object *) 21;
                                        r1 = box(unbox(r1)*unbox(r2));
                                        r2 = r1;
                                        r3 = r0;
                                        r1 = box(unbox(r2)+unbox(r3));
                                        r2 = varref (1,0);
                                        r3 = (object *) 3;
                                        r2 = box(unbox(r2)+unbox(r3));
                                        lenv = ((object *)lenv)[1];
                                        lenv[2] = r2;
                                        lenv[3] = r1;
                                        goto FUN_loop_626_i135;
                                      } break;
                                      default: {
                                        r0 = (object*) &constructed_66;
                                        dump_object (r0, 0); fprintf (stdout, "\n");
                                        goto Lreturn;
                                        r0 = PXLL_UNDEFINED;
                                        PXLL_RETURN(0);
                                      } break;
                                    }
                                    PXLL_RETURN(0);
                                  }
                                  PXLL_RETURN(0);
                                L121:
                                r1 = allocate (TC_CLOSURE, 2);
                                r1[1] = &&FUN_loop_626_i135; r1[2] = lenv;
                                r0[2] = r1;
                                r0 = allocate (TC_TUPLE, 3);
                                r1 = (object *) 1;
                                r0[2] = r1;
                                r1 = (object *) 1;
                                r0[3] = r1;
                                r1 = varref (0,0);
                                r0[1] = r1[2]; lenv = r0; goto FUN_loop_626_i135;
                              } break;
                            }
                            PXLL_RETURN(0);
                          } break;
                        }
                        PXLL_RETURN(0);
                      }
                      PXLL_RETURN(0);
                    }
                    PXLL_RETURN(0);
                  }
                  PXLL_RETURN(0);
                }
                PXLL_RETURN(0);
              }
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L75:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read1_554; r1[2] = lenv;
    r0[18] = r1;
    // def 'read_555'
    goto L123;
    FUN_read_555:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (1,16);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L124; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_read1_554;
      L124:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (1,14);
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L125;  k = t;
      lenv = r0[2]; goto FUN_skip_whitespace_552;
      L125:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (2,1);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L126; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_peek_548;
      L126:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 23298;
      r1 = varref (0,0);
      if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (3,18);
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L127; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_read_array_index_556;
        L127:
        r0 = k[4];; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = varref (0,0);
        r1 = (object *) TC_NIL;
        t = alloc_no_clear (TC_PAIR, 2);
        t[1] = r0; t[2] = r1;
        r0 = t;
        r1 = varref (2,0);
        t = alloc_no_clear (TC_PAIR, 2);
        t[1] = r1; t[2] = r0;
        r0 = t;
        r1 = (object*) constructed_51[0];
        t = alloc_no_clear ((TC_USEROBJ+4), 1);
        t[1] = r1;
        r1 = t;
        t = alloc_no_clear (TC_PAIR, 2);
        t[1] = r1; t[2] = r0;
        r0 = t;
        t = alloc_no_clear ((TC_USEROBJ+0), 1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } else {
        r0 = (object *) 14850;
        r1 = varref (0,0);
        if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (4,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (5,0);
          r0[2] = r1;
          r1 = top[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L128;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
          L128:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          varset (4, 0, r0);
          r0 = varref (0,0);
          lenv = lenv[1];
          r0 = varref (2,17);
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L129;  k = t;
          lenv = r0[2]; goto FUN_read_555;
          L129:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          r1 = r0;
          r2 = (object *) TC_NIL;
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r1; t[2] = r2;
          r1 = t;
          r2 = varref (1,0);
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r2; t[2] = r1;
          r1 = t;
          r2 = (object*) constructed_22[0];
          t = alloc_no_clear ((TC_USEROBJ+4), 1);
          t[1] = r2;
          r2 = t;
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r2; t[2] = r1;
          r1 = t;
          t = alloc_no_clear ((TC_USEROBJ+0), 1);
          t[1] = r1;
          r1 = t;
          PXLL_RETURN(1);
        } else {
          r0 = varref (1,0);
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L123:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_555; r1[2] = lenv;
    r0[19] = r1;
    // def 'read-array-index_556'
    goto L130;
    FUN_read_array_index_556:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (2,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (3,0);
      r0[2] = r1;
      r1 = top[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L131;  k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
      L131:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      varset (2, 0, r0);
      r0 = varref (0,0);
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (1,17);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L132; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_read_555;
      L132:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (1,14);
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L133;  k = t;
      lenv = r0[2]; goto FUN_skip_whitespace_552;
      L133:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (2,1);
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L134; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_peek_548;
      L134:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = varref (0,0);
      r1 = (object *) 23810;
      r0 = PXLL_TEST(r0==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (3,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (4,0);
        r0[2] = r1;
        r1 = top[4];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L135;  k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
        L135:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        varset (3, 0, r0);
        r0 = varref (0,0);
        lenv = lenv[1];
        r0 = varref (0,0);
        PXLL_RETURN(0);
      } else {
        r0 = (object*) &constructed_49;
        dump_object (r0, 0); fprintf (stdout, "\n");
        goto Lreturn;
        r0 = PXLL_UNDEFINED;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L130:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_array_index_556; r1[2] = lenv;
    r0[20] = r1;
    // def 'read-list_560'
    goto L136;
    FUN_read_list_560:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (2,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (3,0);
      r0[2] = r1;
      r1 = top[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L137;  k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
      L137:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      varset (2, 0, r0);
      r0 = varref (0,0);
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def 'loop_612'
      goto L138;
      FUN_loop_612:
        check_heap (0);
        r0 = varref (2,14);
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L139;  k = t;
        lenv = r0[2]; goto FUN_skip_whitespace_552;
        L139:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (3,1);
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L140; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_peek_548;
        L140:
        r0 = k[4];; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = varref (0,0);
        r1 = (object *) 10498;
        if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (5,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (6,0);
          r0[2] = r1;
          r1 = top[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L141;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
          L141:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          varset (5, 0, r0);
          r0 = varref (0,0);
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = varref (1,0);
          r0[2] = r1;
          r1 = (object *) TC_NIL;
          r0[3] = r1;
          r1 = top[5];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_69;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (4,17);
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L142; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_read_555;
          L142:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = varref (0,0);
          r1 = varref (2,0);
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r0; t[2] = r1;
          r0 = t;
          lenv = ((object **)lenv)[1][1];
          lenv[2] = r0;
          goto FUN_loop_612;
        }
        PXLL_RETURN(0);
      L138:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_612; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) constructed_48[0];
      r0[2] = r1;
      r1 = varref (0,0);
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_612;
    L136:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_list_560; r1[2] = lenv;
    r0[21] = r1;
    // def 'read-record_561'
    goto L143;
    FUN_read_record_561:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = varref (2,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (3,0);
      r0[2] = r1;
      r1 = top[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L144;  k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
      L144:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      varset (2, 0, r0);
      r0 = varref (0,0);
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def 'loop_616'
      goto L145;
      FUN_loop_616:
        check_heap (0);
        r0 = varref (2,14);
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L146;  k = t;
        lenv = r0[2]; goto FUN_skip_whitespace_552;
        L146:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (3,1);
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L147; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_peek_548;
        L147:
        r0 = k[4];; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = varref (0,0);
        r1 = (object *) 32002;
        if PXLL_IS_TRUE(PXLL_TEST(r0==r1)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (5,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (6,0);
          r0[2] = r1;
          r1 = top[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L148;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
          L148:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          varset (5, 0, r0);
          r0 = varref (0,0);
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = varref (1,0);
          r0[2] = r1;
          r1 = (object *) TC_NIL;
          r0[3] = r1;
          r1 = top[5];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L149;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_69;
          L149:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          t = alloc_no_clear ((TC_USEROBJ+32), 1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          // def 'loop_621_i178'
          goto L150;
          FUN_loop_621_i178:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (0,1);
            r0[2] = r1;
            r1 = varref (6,13);
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L151;  k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L151:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = varref (8,0);
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r1 = varref (9,0);
              r0[2] = r1;
              r1 = top[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L152;  k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
              L152:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              varset (8, 0, r0);
              r0 = varref (0,0);
              lenv = lenv[1];
              r0 = varref (6,1);
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L153;  k = t;
              lenv = r0[2]; goto FUN_peek_548;
              L153:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              r1 = varref (0,1);
              r2 = varref (0,0);
              t = alloc_no_clear (TC_PAIR, 2);
              t[1] = r1; t[2] = r2;
              r1 = t;
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_loop_621_i178;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 3);
              r2 = varref (0,0);
              r1[2] = r2;
              r2 = (object *) TC_NIL;
              r1[3] = r2;
              r2 = top[5];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L154; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
              L154:
              r0 = k[4];; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = top[8];
              r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_97;
            }
            PXLL_RETURN(0);
          L150:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN_loop_621_i178; r2[2] = lenv;
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) constructed_47[0];
          r1[2] = r2;
          r2 = varref (5,1);
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L155; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_peek_548;
          L155:
          r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = varref (0,0);
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L156; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_loop_621_i178;
          L156:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          r0[2] = r1;
          r0 = varref (4,14);
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L157;  k = t;
          lenv = r0[2]; goto FUN_skip_whitespace_552;
          L157:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = varref (6,1);
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L158; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_peek_548;
          L158:
          r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = varref (0,0);
          r2 = (object *) 15618;
          r1 = PXLL_TEST(r1==r2);
          lenv = lenv[1];
          r0[2] = r1;
          r1 = (object *) PXLL_FALSE;
          r0[3] = r1;
          r0 = varref (0,0);
          r1 = varref (0,1);
          r0 = PXLL_TEST(r0==r1);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object*) &constructed_46;
            dump_object (r0, 0); fprintf (stdout, "\n");
            goto Lreturn;
            r0 = PXLL_UNDEFINED;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = varref (6,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = varref (7,0);
            r0[2] = r1;
            r1 = top[4];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L159;  k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_115;
            L159:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            varset (6, 0, r0);
            r0 = varref (0,0);
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = varref (5,17);
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L160; t[4] = r0; k = t;
            lenv = r1[2]; goto FUN_read_555;
            L160:
            r0 = k[4];; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = varref (1,0);
            r1 = varref (0,0);
            t = alloc_no_clear ((TC_USEROBJ+0), 2);
            t[1] = r0; t[2] = r1;
            r0 = t;
            r1 = varref (3,0);
            t = alloc_no_clear (TC_PAIR, 2);
            t[1] = r0; t[2] = r1;
            r0 = t;
            lenv = ((object ***)lenv)[1][1][1];
            lenv[2] = r0;
            goto FUN_loop_616;
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L145:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_616; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) constructed_45[0];
      r0[2] = r1;
      r1 = varref (0,0);
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_616;
    L143:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_record_561; r1[2] = lenv;
    r0[22] = r1;
    // def 'read-include_564'
    goto L161;
    FUN_read_include_564:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = varref (1,0);
      r1[2] = r2;
      r2 = top[17];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L162; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_read_file_133;
      L162:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = varref (0,0);
      r1[2] = r2;
      r2 = (object *) TC_NIL;
      r1[3] = r2;
      r2 = top[5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L163; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
      L163:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = allocate (TC_TUPLE, 3);
      r2 = varref (0,1);
      r1[2] = r2;
      r2 = (object *) TC_NIL;
      r1[3] = r2;
      r2 = top[5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L164; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_69;
      L164:
      r0 = k[4];; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = varref (1,1);
      r0[3] = r1;
      r1 = top[5];
      r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_69;
    L161:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_include_564; r1[2] = lenv;
    r0[23] = r1;
    // def 'read-all_565'
    goto L165;
    FUN_read_all_565:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def 'loop_633'
      goto L166;
      FUN_loop_633:
        check_heap (0);
        r0 = varref (2,14);
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L167;  k = t;
        lenv = r0[2]; goto FUN_skip_whitespace_552;
        L167:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (3,1);
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L168; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_peek_548;
        L168:
        r0 = k[4];; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = varref (0,0);
        r1 = (object *) 65794;
        r0 = PXLL_TEST(r0==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 3);
          r1 = varref (0,0);
          r0[2] = r1;
          r1 = (object *) TC_NIL;
          r0[3] = r1;
          r1 = top[5];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_69;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = varref (3,17);
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L169; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_read_555;
          L169:
          r0 = k[4];; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = varref (0,0);
          switch (get_case (r0)) {
            case (TC_USEROBJ+0): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = varref (1,0);
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = varref (0,0);
              switch (get_case_noint (r0)) {
                case (((pxll_int)TC_PAIR)): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = varref (1,0);
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = varref (1,0);
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = varref (0,0);
                  switch (get_case (r0)) {
                    case (TC_USEROBJ+4): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = varref (1,0);
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = (object*) constructed_44[0];
                      r1 = r0;
                      r2 = varref (0,0);
                      r0 = PXLL_TEST(r1==r2);
                      if PXLL_IS_TRUE(r0) {
                        r0 = varref (1,1);
                        switch (get_case_noint (r0)) {
                          case (((pxll_int)TC_PAIR)): {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = varref (2,1);
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = varref (2,1);
                            r1 = UOBJ_GET(r1,1);
                            r0[3] = r1;
                            r0 = varref (0,0);
                            switch (get_case (r0)) {
                              case (TC_STRING): {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = varref (1,0);
                                r0[2] = r1;
                                r0 = varref (1,1);
                                switch (get_case_noint (r0)) {
                                  case (((pxll_int)TC_NIL)): {
                                    r0 = allocate (TC_TUPLE, 3);
                                    r1 = varref (0,0);
                                    r0[2] = r1;
                                    r1 = varref (6,0);
                                    r0[3] = r1;
                                    r1 = varref (8,21);
                                    t = allocate (TC_SAVE, 3);
                                    t[1] = k; t[2] = lenv; t[3] = &&L170;  k = t;
                                    r0[1] = r1[2]; lenv = r0; goto FUN_read_include_564;
                                    L170:
                                    ; lenv = k[2]; k = k[1];
                                    r0 = result;
                                    lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                    lenv[2] = r0;
                                    goto FUN_loop_633;
                                  } break;
                                  default: {
                                    lenv = ((object *****)lenv)[1][1][1][1][1];
                                    goto fatbar_2;
                                  }
                                }
                                PXLL_RETURN(0);
                              } break;
                              default: {
                                lenv = ((object ****)lenv)[1][1][1][1];
                                goto fatbar_2;
                              }
                            }
                            PXLL_RETURN(0);
                          } break;
                          default: {
                            lenv = ((object ***)lenv)[1][1][1];
                            goto fatbar_2;
                          }
                        }
                        PXLL_RETURN(0);
                      } else {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_2;
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_2;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_2;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_2;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_2_over;
          fatbar_2:
          r0 = varref (0,0);
          r1 = varref (1,0);
          t = alloc_no_clear (TC_PAIR, 2);
          t[1] = r0; t[2] = r1;
          r0 = t;
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_loop_633;
          fatbar_2_over: ;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L166:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_633; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) constructed_42[0];
      r0[2] = r1;
      r1 = varref (0,0);
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_633;
    L165:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_all_565; r1[2] = lenv;
    r0[24] = r1;
    r0 = varref (0,22);
    lenv = r0[2]; goto FUN_read_all_565;
  L33:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_reader_132; r1[2] = lenv;
  r0[16] = r1;
  // def 'read-file_133'
  goto L171;
  FUN_read_file_133:
    check_heap (0);
    r0 = (object*) &constructed_41;
    r0 = box(((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_40;
    box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
    r0 = varref (0,0);
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_USEROBJ+4, 4);
    r2 = (object *) 32769;
    r3 = (object *) 32769;
    r2 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r2))), ((pxll_string*)(t))->len = unbox(r3), t);
    r1[1] = r2;
    r2 = (object *) 1;
    r1[2] = r2;
    r2 = allocate (TC_TUPLE, 4);
    r3 = varref (1,0);
    r2[2] = r3;
    r3 = top[15];
    r2[3] = r3;
    r3 = (object *) 1;
    r2[4] = r3;
    r3 = top[12];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L172; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_open_101;
    L172:
    r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = (object *) 1;
    r1[4] = r2;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = varref (0,0);
    r0[2] = r1;
    r1 = top[16];
    r0[1] = r1[2]; lenv = r0; goto FUN_reader_132;
  L171:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_read_file_133; r1[2] = lenv;
  r0[17] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  r2 = box(argc);
  r1[2] = r2;
  r2 = varref (0,0);
  r3 = (object*) &constructed_39;
  if (unbox(r2) == 0) { r2 = (object *) TC_EMPTY_VECTOR; } else {
    t = alloc_no_clear (TC_VECTOR, unbox(r2));
    for (i=0; i < unbox(r2); i++) { t[i+1] = r3; }
    r2 = t;
  }
  r1[3] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r2[1] = lenv; lenv = r2;
  // def 'loop_683'
  goto L173;
  FUN_loop_683:
    r0 = varref (0,0);
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r0)==0)) {
      r0 = varref (2,1);
      PXLL_RETURN(0);
    } else {
      r0 = varref (0,0);
      r1 = (object *) 3;
      r0 = box(unbox(r0)-unbox(r1));
      r1 = r0;
      r1 = box(strlen(argv[unbox(r1)]));
      r2 = r1;
      r3 = r1;
      r2 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r2))), ((pxll_string*)(t))->len = unbox(r3), t);
      r3 = r2;
      r4 = r0;
      r5 = r1;
      (memcpy (((pxll_string*)(r3))->data, argv[unbox(r4)], unbox(r5)), PXLL_UNDEFINED);
      r0 = r2;
      r1 = varref (0,0);
      r2 = (object *) 3;
      r1 = box(unbox(r1)-unbox(r2));
      r2 = varref (2,1);
      range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
      ((pxll_vector*)r2)->val[unbox(r1)] = r0;
      r0 = varref (0,0);
      r1 = (object *) 3;
      r0 = box(unbox(r0)-unbox(r1));
      lenv[2] = r0;
      goto FUN_loop_683;
    }
    PXLL_RETURN(0);
  L173:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_loop_683; r3[2] = lenv;
  r2[2] = r3;
  r2 = allocate (TC_TUPLE, 2);
  r3 = varref (1,0);
  r2[2] = r3;
  r3 = varref (0,0);
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L174; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_loop_683;
  L174:
  r0 = k[4]; r1 = k[5];; lenv = k[2]; k = k[1];
  r2 = result;
  lenv = lenv[1];
  r1[4] = r2;
  r1 = allocate (TC_USEROBJ+0, 2);
  r2 = varref (0,0);
  r1[1] = r2;
  r2 = varref (0,2);
  r1[2] = r2;
  lenv = lenv[1];
  r0[18] = r1;
  // def 'for-each_79'
  goto L175;
  FUN_for_each_79:
    check_heap (0);
    r0 = varref (0,1);
    switch (get_case_noint (r0)) {
      case (((pxll_int)TC_PAIR)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,1);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = varref (1,1);
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (0,0);
        r0[2] = r1;
        r1 = varref (1,0);
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L176;  k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L176:
        ; lenv = k[2]; k = k[1];
        r0 = varref (1,0);
        r1 = varref (0,1);
        lenv = ((object *)lenv)[1];
        lenv[2] = r0;
        lenv[3] = r1;
        goto FUN_for_each_79;
      } break;
      default: {
        r0 = (object *) PXLL_FALSE;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L175:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_for_each_79; r1[2] = lenv;
  r0[19] = r1;
  // def 'unread-fields_135'
  goto L177;
  FUN_unread_fields_135:
    check_heap (0);
    r0 = varref (0,0);
    switch (get_case_noint (r0)) {
      case (((pxll_int)TC_PAIR)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = varref (0,0);
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = varref (0,0);
        r0 = box(((pxll_string*)(r0))->len);
        r1 = varref (0,0);
        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        r0 = (object*) &constructed_38;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_37;
        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (0,1);
        r0[2] = r1;
        r1 = top[22];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L178;  k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
        L178:
        ; lenv = k[2]; k = k[1];
        r0 = (object*) &constructed_36;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_35;
        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        r0 = varref (1,1);
        lenv = ((object **)lenv)[1][1];
        lenv[2] = r0;
        goto FUN_unread_fields_135;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L177:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unread_fields_135; r1[2] = lenv;
  r0[20] = r1;
  // def 'unread-list_136'
  goto L179;
  FUN_unread_list_136:
    check_heap (0);
    r0 = varref (0,0);
    switch (get_case_noint (r0)) {
      case (((pxll_int)TC_PAIR)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = varref (0,0);
        switch (get_case (r0)) {
          case (TC_USEROBJ+4): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = varref (1,0);
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = (object*) constructed_28[0];
            r1 = r0;
            r2 = varref (0,0);
            r0 = PXLL_TEST(r1==r2);
            if PXLL_IS_TRUE(r0) {
              r0 = varref (1,1);
              switch (get_case_noint (r0)) {
                case (((pxll_int)TC_PAIR)): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = varref (2,1);
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = varref (2,1);
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = varref (0,1);
                  switch (get_case_noint (r0)) {
                    case (((pxll_int)TC_NIL)): {
                      r0 = (object*) &constructed_16;
                      r0 = box(((pxll_string*)(r0))->len);
                      r1 = (object*) &constructed_15;
                      box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = varref (0,0);
                      r0[2] = r1;
                      r1 = top[22];
                      r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_1;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_1;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_1_over;
              fatbar_1:
              r0 = (object*) constructed_22[0];
              r1 = r0;
              r2 = varref (0,0);
              r0 = PXLL_TEST(r1==r2);
              if PXLL_IS_TRUE(r0) {
                r0 = varref (1,1);
                switch (get_case_noint (r0)) {
                  case (((pxll_int)TC_PAIR)): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (2,1);
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = varref (2,1);
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = varref (0,1);
                    switch (get_case_noint (r0)) {
                      case (((pxll_int)TC_PAIR)): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = varref (1,1);
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = varref (1,1);
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = varref (0,1);
                        switch (get_case_noint (r0)) {
                          case (((pxll_int)TC_NIL)): {
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = varref (1,0);
                            r0[2] = r1;
                            r1 = top[22];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L180;  k = t;
                            r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                            L180:
                            ; lenv = k[2]; k = k[1];
                            r0 = (object*) &constructed_18;
                            r0 = box(((pxll_string*)(r0))->len);
                            r1 = (object*) &constructed_17;
                            box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = varref (0,0);
                            r0[2] = r1;
                            r1 = top[22];
                            r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                          } break;
                          default: {
                            lenv = ((object ****)lenv)[1][1][1][1];
                            goto fatbar_0;
                          }
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        r0 = (object*) &constructed_20;
                        r0 = box(((pxll_string*)(r0))->len);
                        r1 = (object*) &constructed_19;
                        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = varref (0,0);
                        r0[2] = r1;
                        r1 = top[22];
                        r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                      } break;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_0;
                  }
                }
                PXLL_RETURN(0);
              } else {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_0;
              }
              PXLL_RETURN(0);
              fatbar_1_over: ;
              PXLL_RETURN(0);
            } else {
              r0 = (object*) constructed_22[0];
              r1 = r0;
              r2 = varref (0,0);
              r0 = PXLL_TEST(r1==r2);
              if PXLL_IS_TRUE(r0) {
                r0 = varref (1,1);
                switch (get_case_noint (r0)) {
                  case (((pxll_int)TC_PAIR)): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = varref (2,1);
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = varref (2,1);
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = varref (0,1);
                    switch (get_case_noint (r0)) {
                      case (((pxll_int)TC_PAIR)): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = varref (1,1);
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = varref (1,1);
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = varref (0,1);
                        switch (get_case_noint (r0)) {
                          case (((pxll_int)TC_NIL)): {
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = varref (1,0);
                            r0[2] = r1;
                            r1 = top[22];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L181;  k = t;
                            r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                            L181:
                            ; lenv = k[2]; k = k[1];
                            r0 = (object*) &constructed_24;
                            r0 = box(((pxll_string*)(r0))->len);
                            r1 = (object*) &constructed_23;
                            box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = varref (0,0);
                            r0[2] = r1;
                            r1 = top[22];
                            r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                          } break;
                          default: {
                            lenv = ((object ****)lenv)[1][1][1][1];
                            goto fatbar_0;
                          }
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        r0 = (object*) &constructed_26;
                        r0 = box(((pxll_string*)(r0))->len);
                        r1 = (object*) &constructed_25;
                        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = varref (0,0);
                        r0[2] = r1;
                        r1 = top[22];
                        r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
                      } break;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_0;
                  }
                }
                PXLL_RETURN(0);
              } else {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_0;
              }
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_0;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_0;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_0_over;
    fatbar_0:
    r0 = (object*) &constructed_34;
    r0 = box(((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_33;
    box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
    r0 = allocate (TC_TUPLE, 3);
    // def None
    goto L182;
    FUN_17259_lambda:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (0,0);
      r0[2] = r1;
      r1 = top[22];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L183;  k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
      L183:
      ; lenv = k[2]; k = k[1];
      r0 = (object*) &constructed_32;
      r0 = box(((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_31;
      r0 = box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
      PXLL_RETURN(0);
    L182:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_17259_lambda; r1[2] = lenv;
    r0[2] = r1;
    r1 = varref (0,0);
    r0[3] = r1;
    r1 = top[19];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L184;  k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_79;
    L184:
    ; lenv = k[2]; k = k[1];
    r0 = (object*) &constructed_30;
    r0 = box(((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_29;
    r0 = box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
    PXLL_RETURN(0);
    fatbar_0_over: ;
    PXLL_RETURN(0);
  L179:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unread_list_136; r1[2] = lenv;
  r0[21] = r1;
  // def 'unread_137'
  goto L185;
  FUN_unread_137:
    check_heap (0);
    r0 = varref (0,0);
    switch (get_case (r0)) {
      case (TC_UNDEFINED): {
        r0 = (object *) 14;
        dump_object (r0, 0);
        r0 = (object *) 1;
        PXLL_RETURN(0);
      } break;
      case (TC_STRING): {
        r0 = varref (0,0);
        r1 = r0;
        dump_object (r1, 0);
        r1 = (object *) 1;
        PXLL_RETURN(1);
      } break;
      case (TC_INT): {
        r0 = varref (0,0);
        r1 = r0;
        dump_object (r1, 0);
        r1 = (object *) 1;
        PXLL_RETURN(1);
      } break;
      case (TC_USEROBJ+4): {
        r0 = varref (0,0);
        r0 = UOBJ_GET(r0,0);
        r1 = r0;
        dump_object (r1, 0);
        r1 = (object *) 1;
        PXLL_RETURN(1);
      } break;
      case (TC_USEROBJ+0): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (0,0);
        r0[2] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_unread_list_136;
      } break;
      case (TC_CHAR): {
        r0 = varref (0,0);
        r1 = r0;
        dump_object (r1, 0);
        r1 = (object *) 1;
        PXLL_RETURN(1);
      } break;
      case (TC_USEROBJ+32): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = (object*) &constructed_4;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_3;
        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        r0 = allocate (TC_TUPLE, 2);
        r1 = varref (0,0);
        r0[2] = r1;
        r1 = top[20];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L186;  k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_unread_fields_135;
        L186:
        ; lenv = k[2]; k = k[1];
        r0 = (object*) &constructed_2;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_1;
        r0 = box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        PXLL_RETURN(0);
      } break;
      case (TC_USEROBJ+28): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = varref (1,0);
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = (object*) &constructed_10;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_9;
        box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        r0 = allocate (TC_TUPLE, 3);
        // def None
        goto L187;
        FUN_17346_lambda:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r1 = varref (0,0);
          r0[2] = r1;
          r1 = top[22];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L188;  k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
          L188:
          ; lenv = k[2]; k = k[1];
          r0 = (object*) &constructed_8;
          r0 = box(((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_7;
          r0 = box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
          PXLL_RETURN(0);
        L187:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_17346_lambda; r1[2] = lenv;
        r0[2] = r1;
        r1 = varref (0,0);
        r0[3] = r1;
        r1 = top[19];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L189;  k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_79;
        L189:
        ; lenv = k[2]; k = k[1];
        r0 = (object*) &constructed_6;
        r0 = box(((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_5;
        r0 = box(fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout));
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = varref (0,0);
        r0 = UOBJ_GET(r0,0);
        r1 = r0;
        switch ((pxll_int) (r1)) {
          case (((pxll_int)PXLL_FALSE)): {
            r1 = (object*) &constructed_12;
            r1 = box(((pxll_string*)(r1))->len);
            r2 = (object*) &constructed_11;
            r1 = box(fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout));
            PXLL_RETURN(1);
          } break;
          default: {
            r1 = (object*) &constructed_14;
            r1 = box(((pxll_string*)(r1))->len);
            r2 = (object*) &constructed_13;
            r1 = box(fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout));
            PXLL_RETURN(1);
          } break;
        }
        PXLL_RETURN(1);
      } break;
    }
    PXLL_RETURN(0);
  L185:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unread_137; r1[2] = lenv;
  r0[22] = r1;
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = (object *) pxll_internal_symbols;
  r0[2] = r1;
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  // def 'loop_536_i237'
  goto L190;
  FUN_loop_536_i237:
    check_heap (0);
    r0 = varref (2,0);
    r1 = varref (2,0);
    r0 = box((r0 == (object*) TC_EMPTY_VECTOR) ? 0 : GET_TUPLE_LENGTH(*r1));
    r1 = varref (0,0);
    r2 = r0;
    r0 = PXLL_TEST(unbox(r1)==unbox(r2));
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r1 = varref (2,0);
      r2 = varref (0,0);
      range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r2));
      r1 = ((pxll_vector*)r1)->val[unbox(r2)];
      r0[2] = r1;
      r1 = top[14];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L191;  k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_intern_symbol_127;
      L191:
      ; lenv = k[2]; k = k[1];
      r0 = varref (0,0);
      r1 = (object *) 3;
      r0 = box(unbox(r0)+unbox(r1));
      lenv[2] = r0;
      goto FUN_loop_536_i237;
    }
    PXLL_RETURN(0);
  L190:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_loop_536_i237; r1[2] = lenv;
  r0[2] = r1;
  r0 = allocate (TC_TUPLE, 2);
  r1 = (object *) 1;
  r0[2] = r1;
  r1 = varref (0,0);
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L192;  k = t;
  r0[1] = r1[2]; lenv = r0; goto FUN_loop_536_i237;
  L192:
  ; lenv = k[2]; k = k[1];
  r0 = result;
  lenv = lenv[1];
  lenv = lenv[1];
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = allocate (TC_TUPLE, 2);
  r2 = top[18];
  r2 = ((pxll_vector*)r2)->val[0];
  r3 = r2;
  r4 = (object *) 3;
  r2 = PXLL_TEST(unbox(r3)>unbox(r4));
  if PXLL_IS_TRUE(r2) {
    r2 = top[18];
    r2 = ((pxll_vector*)r2)->val[1];
    r3 = (object *) 3;
    range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r3));
    r2 = ((pxll_vector*)r2)->val[unbox(r3)];
  } else {
    r2 = (object*) &constructed_0;
  }
  r1[2] = r2;
  r2 = top[17];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L193; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_read_file_133;
  L193:
  r0 = k[4];; lenv = k[2]; k = k[1];
  r1 = result;
  r0[2] = r1;
  r0 = allocate (TC_TUPLE, 3);
  // def None
  goto L194;
  FUN_17503_lambda:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = varref (0,0);
    r0[2] = r1;
    r1 = top[22];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L195;  k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_unread_137;
    L195:
    ; lenv = k[2]; k = k[1];
    r0 = (object *) 2562;
    r0 = box(fputc (GET_CHAR(r0), stdout));
    PXLL_RETURN(0);
  L194:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_17503_lambda; r1[2] = lenv;
  r0[2] = r1;
  r1 = varref (0,0);
  r0[3] = r1;
  r1 = top[19];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L196;  k = t;
  r0[1] = r1[2]; lenv = r0; goto FUN_for_each_79;
  L196:
  ; lenv = k[2]; k = k[1];
  r0 = (object *) 14;
  PXLL_RETURN(0);
 Lreturn:
 return (pxll_int) result;
}
static int lookup_field (int tag, int label)
{  switch (tag) {
  case 1:
  switch (label) {
    case 2: return 0; break;
    case 3: return 1; break;
    case 4: return 2; break;
    case 5: return 3; break;
  } break;
  case 0:
  switch (label) {
    case 0: return 0; break;
    case 1: return 1; break;
  } break;
}}
